from rpcclient.exceptions import SymbolAbsentError


class SymbolsJar(dict):
    @staticmethod
    def create(client):
        """
        Factory method for creating symbols jars
        :param client: client
        :rtype: SymbolsJar
        """
        jar = SymbolsJar()
        jar.__dict__['_client'] = client
        return jar

    def get_lazy(self, name):
        client = self.__dict__['_client']
        s = client.dlsym(client._dlsym_global_handle, name)
        if 0 == s:
            raise SymbolAbsentError(f'no such loaded symbol: {name}')
        self[name] = client.symbol(s)
        return self[name]

    def __getitem__(self, item):
        if item not in self:
            symbol = self.get_lazy(item)
            if symbol:
                return symbol
        return dict.__getitem__(self, item)

    def __getattr__(self, name):
        if name not in self:
            client = self.__dict__['_client']
            s = client.dlsym(client._dlsym_global_handle, name)
            if 0 == s:
                raise SymbolAbsentError(f'no such loaded symbol: {name}')
            self[name] = client.symbol(s)
            return self[name]

        return self.get(name)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def __delattr__(self, item):
        return self.__delitem__(item)

    def __sub__(self, other):
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k1, v1 in self.items():
            if k1 not in other:
                retval[k1] = v1
        return retval

    def __add__(self, other):
        retval = SymbolsJar.create(self.__dict__['_client'])
        for k, v in other.items():
            retval[k] = v
        for k, v in self.items():
            retval[k] = v
        return retval



from collections import namedtuple

import lief
from parameter_decorators import path_to_str

Symbol = namedtuple('Symbol', 'origin value')


class Lief:
    """" parse and patch executable files """

    def __init__(self, client):
        self._client = client

    @path_to_str('path')
    def parse(self, path: str):
        with self._client.fs.open(path, 'r') as f:
            return lief.parse(f.read())

    @path_to_str('path')
    def get_symbols(self, path: str) -> dict[str, Symbol]:
        result = {}
        parsed = self.parse(path)
        for s in parsed.symbols:
            result[s.name] = Symbol(origin=s.origin, value=s.value)
        return result



from typing import Optional, Tuple

from rpcclient.darwin.structs import POLLIN, pollfd
from rpcclient.network import AF_UNIX, SOCK_STREAM, Socket

FD_SIZE = 4
READ_SIZE = 0x10000


class CaptureFD:
    """
    Context manager, capturing output to a given `fd`. Read from it using the `read()` method.
    """

    def __init__(self, client, fd: int, sock_buf_size: Optional[int] = None) -> None:
        """
        sock_buf_size is required for captures above 6KB, as any write above this value would block until a read is performed.

        :param rpcclient.client.client.Client client: Current client
        :param fd: FD to capture
        :param sock_buf_size: Buffer size for the capture socket, if not specified, default value is used.
        """
        self._client = client
        self.fd: int = fd
        self._backupfd: Optional[int] = None
        self._socket_pair: Optional[Tuple[int, int]] = None
        self._sock_buf_size: Optional[int] = sock_buf_size

    def __enter__(self) -> 'CaptureFD':
        with self._client.safe_malloc(FD_SIZE * 2) as socket_pair:
            socket_pair.item_size = FD_SIZE
            if 0 != self._client.symbols.socketpair(AF_UNIX, SOCK_STREAM, 0, socket_pair):
                self._client.raise_errno_exception('socketpair failed')
            self._socket_pair = (socket_pair[0].c_int32, socket_pair[1].c_int32)
        if self._sock_buf_size is not None:
            Socket(self._client, self._socket_pair[0]).setbufsize(self._sock_buf_size)
        self._backupfd = self._client.symbols.dup(self.fd).c_int32
        if -1 == self._backupfd:
            self._backupfd = None
            self._client.raise_errno_exception('dup fd failed')
        if 0 > self._client.symbols.dup2(self._socket_pair[0], self.fd):
            self._client.raise_errno_exception('dup2 sock-fd failed')
        return self

    def __exit__(self, type, value, traceback) -> None:
        if self._backupfd is not None:
            if 0 > self._client.symbols.dup2(self._backupfd, self.fd):
                self._client.raise_errno_exception('dup2 backup-fd failed')
            if 0 != self._client.symbols.close(self._backupfd):
                self._client.raise_errno_exception('close backupfd failed')
            self._backupfd = None
        if self._socket_pair is not None:
            if 0 != self._client.symbols.close(self._socket_pair[0]):
                self._client.raise_errno_exception(
                    f'close _socket_pair[0] {self._socket_pair[0]} failed')
            if 0 != self._client.symbols.close(self._socket_pair[1]):
                self._client.raise_errno_exception(
                    f'close _socket_pair[1] {self._socket_pair[1]} failed')
            self._socket_pair = None

    def read(self) -> bytes:
        """ Read the bytes captured from `fd` so far. """
        data = b''
        if self._socket_pair is not None:
            with self._client.safe_malloc(READ_SIZE) as buff:
                read = READ_SIZE
                while read == READ_SIZE:
                    with self._client.safe_malloc(pollfd.sizeof()) as pfds:
                        pfds.poke(pollfd.build({'fd': self._socket_pair[1], 'events': POLLIN, 'revents': 0}))
                        if 1 != self._client.symbols.poll(pfds, 1, 0):
                            return data
                    read = self._client.symbols.read(
                        self._socket_pair[1],
                        buff,
                        READ_SIZE).c_int32
                    if -1 == read:
                        self._client.raise_errno_exception('read fd failed')
                    data += buff.peek(read)
        return data



from enum import Enum
from typing import Optional

from rpcclient.exceptions import MissingLibraryError, RpcPermissionError
from rpcclient.structs.consts import RTLD_NOW


class CLAuthorizationStatus(Enum):
    kCLAuthorizationStatusNotDetermined = 0
    kCLAuthorizationStatusRestricted = 1
    kCLAuthorizationStatusDenied = 2
    kCLAuthorizationStatusAuthorizedAlways = 3
    kCLAuthorizationStatusAuthorizedWhenInUse = 4
    kCLAuthorizationStatusAuthorized = kCLAuthorizationStatusAuthorizedAlways

    @classmethod
    def from_value(cls, value: int):
        for i in cls:
            if i.value == value:
                return i


class Location:
    """
    Wrapper to CLLocationManager
    https://developer.apple.com/documentation/corelocation/cllocationmanager?language=objc
    """

    def __init__(self, client):
        self._client = client

        self._load_location_library()
        self._CLLocationManager = self._client.symbols.objc_getClass('CLLocationManager')
        self._location_manager = self._CLLocationManager.objc_call('sharedManager')

    def _load_location_library(self):
        options = [
            # macOS
            '/System/Library/Frameworks/CoreLocation.framework/Versions/A/CoreLocation',
            # iOS
            '/System/Library/Frameworks/CoreLocation.framework/CoreLocation'
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('CoreLocation library isn\'t available')

    @property
    def location_services_enabled(self) -> bool:
        """ opt-in status for location services """
        return bool(self._location_manager.objc_call('locationServicesEnabled'))

    @location_services_enabled.setter
    def location_services_enabled(self, value: bool):
        """ opt-in status for location services """
        self._CLLocationManager.objc_call('setLocationServicesEnabled:', value)

    @property
    def authorization_status(self) -> CLAuthorizationStatus:
        """ authorization status for current server process of accessing location services """
        return CLAuthorizationStatus.from_value(self._location_manager.objc_call('authorizationStatus'))

    @property
    def last_sample(self) -> Optional[dict]:
        """ last taken location sample (or None if there isn't any) """
        location = self._location_manager.objc_call('location')
        if not location:
            return None
        return location.objc_call('jsonObject').py()

    def request_always_authorization(self) -> None:
        """ Request authorization to always query location data """
        self._location_manager.objc_call('requestAlwaysAuthorization')

    def start_updating_location(self):
        """ request location updates from CLLocationManager """
        if self.authorization_status.value < CLAuthorizationStatus.kCLAuthorizationStatusAuthorizedAlways.value:
            raise RpcPermissionError()
        self._location_manager.objc_call('startUpdatingLocation')

    def stop_updating_location(self):
        """ stop requesting location updates from CLLocationManager """
        self._location_manager.objc_call('stopUpdatingLocation')

    def request_oneshot_location(self):
        """ requests the one-time delivery of the user’s current location """
        if self.authorization_status.value < CLAuthorizationStatus.kCLAuthorizationStatusAuthorizedAlways.value:
            raise RpcPermissionError()
        self._location_manager.objc_call('requestLocation')



import pytest

from rpcclient.darwin.consts import NSStringEncoding

pytestmark = pytest.mark.darwin


def test_method_by_method_name(client):
    NSString = client.objc_get_class('NSString')
    ascii_encoding = NSStringEncoding.NSASCIIStringEncoding
    str1 = NSString.stringWithCString_encoding_('Taylor Swift', ascii_encoding).objc_symbol
    assert str1.cStringUsingEncoding_(ascii_encoding).peek_str() == 'Taylor Swift'
    assert str1.length == len('Taylor Swift')
    assert str1.lowercaseString().objc_symbol.cStringUsingEncoding_(ascii_encoding).peek_str() == 'taylor swift'
    assert str1.uppercaseString().objc_symbol.cStringUsingEncoding_(ascii_encoding).peek_str() == 'TAYLOR SWIFT'


def test_calling_property(client):
    d = client.symbols.objc_getClass('NSMutableDictionary').objc_call('new')
    # call method
    d.objc_call('setObject:forKey:', client.cf('value'), client.cf('key'))
    # call property
    assert '{\n    key = value;\n}' == d.objc_symbol.description.py()


def test_set_implementation(client):
    pid = client.symbols.getpid()

    client.objc_get_class('NSJSONSerialization').get_method('isValidJSONObject:').set_implementation(
        client.symbols.getpid)
    assert client.objc_get_class('NSJSONSerialization').isValidJSONObject_() == pid


@pytest.mark.parametrize('value', [True, False])
def test_always_return(client, value):
    client.objc_get_class('NSJSONSerialization').get_method('isValidJSONObject:').always_return(value)
    assert client.objc_get_class('NSJSONSerialization').isValidJSONObject_() == value



"""
This module makes sure imports from the *_pb2 modules don't depend on the locally installed protobuf version
"""
import os

os.environ['PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION'] = 'python'
os.environ['TEMPORARILY_DISABLE_PROTOBUF_VERSION_CHECK'] = 'true'

from rpcclient.protos.rpc_pb2 import ARCH_ARM64, Argument, CmdCall, CmdClose, CmdCustom, CmdDlclose, CmdDlopen, \
    CmdDlsym, CmdDummyBlock, CmdExec, CmdGetClassList, CmdListDir, CmdPeek, CmdPoke, CmdShowClass, CmdShowObject, \
    Command, Handshake, Response, ResponseCustom  # noqa: E402

__all__ = ['Argument', 'CmdCall', 'CmdClose', 'CmdDlclose', 'CmdDlopen', 'CmdDlsym', 'CmdDummyBlock', 'CmdExec',
           'CmdGetClassList', 'CmdListDir', 'CmdPeek', 'CmdPoke', 'CmdShowClass', 'CmdShowObject', 'Command',
           'Handshake', 'Response', 'ARCH_ARM64', 'CmdCustom', 'ResponseCustom']



from contextlib import suppress
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import Optional

from pygments import highlight
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import ObjectiveCLexer

from rpcclient.darwin import objc
from rpcclient.darwin.objc import Method
from rpcclient.darwin.objective_c_class import Class
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import RpcClientException
from rpcclient.symbol import Symbol
from rpcclient.symbols_jar import SymbolsJar


class SettingIvarError(RpcClientException):
    """ Raise when trying to set an Ivar too early or when the Ivar doesn't exist. """
    pass


@dataclass
class Ivar:
    name: str
    value: DarwinSymbol
    type_: str
    offset: int


class ObjectiveCSymbol(DarwinSymbol):
    """
    Wrapper object for an objective-c symbol.
    Allowing easier access to its properties, methods and ivars.
    """

    @classmethod
    def create(cls, value: int, client):
        """
        Create an ObjectiveCSymbol object.
        :param value: Symbol address.
        :param rpcclient.darwin.client.Client client: client.
        :return: ObjectiveCSymbol object.
        :rtype: ObjectiveCSymbol
        """
        symbol = super().create(value, client)
        symbol.ivars = []
        symbol.properties = []
        symbol.methods = []
        symbol.class_ = None  # type: Class
        symbol.reload()
        return symbol

    def reload(self):
        """
        Reload object's in-memory layout.
        """
        object_data = self._client.showobject(self)

        ivars_list = [
            Ivar(name=ivar['name'], type_=ivar['type'], offset=ivar['offset'],
                 value=ivar['value']) for ivar in object_data['ivars']
        ]
        methods_list = [
            objc.Method.from_data(method, self._client) for method in object_data['methods']
        ]
        properties_list = [
            objc.Property(name=prop['name'], attributes=objc.convert_encoded_property_attributes(prop['attributes']))
            for prop in object_data['properties']
        ]

        class_object = Symbol.create(object_data['class_address'], self._client)
        class_wrapper = Class(self._client, class_object)

        self.ivars = ivars_list
        self.methods = methods_list
        self.properties = properties_list
        self.class_ = class_wrapper

    def show(self, dump_to: Optional[str] = None, recursive: bool = False):
        """
        Print to terminal the highlighted class description.
        :param dump_to: directory to dump.
        :param recursive: Show methods of super classes.
        """
        formatted = self._to_str(recursive)
        print(highlight(formatted, ObjectiveCLexer(), TerminalTrueColorFormatter(style='native')))

        if dump_to is None:
            return
        (Path(dump_to) / f'{self.class_.name}.m').expanduser().write_text(formatted)

    def objc_call(self, selector: str, *params, **kwargs):
        """
        Make objc_call() from self return ObjectiveCSymbol when it's an objc symbol.
        :param selector: Selector to execute.
        :param params: Additional parameters.
        :return: ObjectiveCSymbol when return type is an objc symbol.
        """
        symbol = super().objc_call(selector, *params, **kwargs)
        try:
            is_objc_type = self.get_method(selector).return_type == 'id'
        except AttributeError:
            is_objc_type = False
        return symbol.objc_symbol if is_objc_type else symbol

    def get_method(self, name: str) -> Method:
        for method in self.methods:
            if method.name == name:
                return method
        raise AttributeError(f'Method "{name}" does not exist')

    def _set_ivar(self, name, value):
        try:
            ivars = self.__getattribute__('ivars')
            class_name = self.__getattribute__('class_').name
        except AttributeError as e:
            raise SettingIvarError from e

        for i, ivar in enumerate(ivars):
            if ivar.name == name:
                size = self.item_size
                if i < len(self.ivars) - 1:
                    size = ivars[i + 1].offset - ivar.offset
                with self.change_item_size(size):
                    self[ivar.offset // size] = value
                    ivar.value = value
                return
        raise SettingIvarError(f'Ivar "{name}" does not exist in "{class_name}"')

    def _to_str(self, recursive=False):
        protocols_buf = f'<{",".join(self.class_.protocols)}>' if self.class_.protocols else ''

        if self.class_.super is not None:
            buf = f'@interface {self.class_.name}: {self.class_.super.name} {protocols_buf}\n'
        else:
            buf = f'@interface {self.class_.name} {protocols_buf}\n'

        # Add ivars
        buf += '{\n'
        for ivar in self.ivars:
            buf += f'\t{ivar.type_} {ivar.name} = 0x{int(ivar.value):x}; // 0x{ivar.offset:x}\n'
        buf += '}\n'

        # Add properties
        for prop in self.properties:
            attrs = prop.attributes
            buf += f'@property ({",".join(attrs.list)}) {prop.attributes.type_} {prop.name};\n'

            if attrs.synthesize is not None:
                buf += f'@synthesize {prop.name} = {attrs.synthesize};\n'

        # Add methods
        methods = self.methods.copy()

        # Add super methods.
        if recursive:
            for sup in self.class_.iter_supers():
                for method in filter(lambda m: m not in methods, sup.methods):
                    methods.append(method)

        # Print class methods first.
        methods.sort(key=lambda m: not m.is_class)

        for method in methods:
            buf += str(method)

        buf += '@end'
        return buf

    @property
    def symbols_jar(self) -> SymbolsJar:
        """ Get a SymbolsJar object for quick operations on all methods """
        jar = SymbolsJar.create(self._client)

        for m in self.methods:
            jar[m.name] = m.address

        return jar

    def __dir__(self):
        result = set()

        for ivar in self.ivars:
            result.add(ivar.name)

        for method in self.methods:
            result.add(method.name.replace(':', '_'))

        for sup in self.class_.iter_supers():
            for method in sup.methods:
                result.add(method.name.replace(':', '_'))

        result.update(list(super().__dir__()))
        return list(result)

    def __getitem__(self, item):
        if isinstance(item, int):
            return super().__getitem__(item)

        # Ivars
        for ivar in self.ivars:
            if ivar.name == item:
                if self._client.is_objc_type(ivar.value):
                    return ivar.value.objc_symbol
                return ivar.value

        # Properties
        for prop in self.properties:
            if prop.name == item:
                return self.objc_call(item)

        # Methods
        for method in self.methods:
            if method.name == item:
                return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        for sup in self.class_.iter_supers():
            for method in sup.methods:
                if method.name == item:
                    return partial(self.class_.objc_call, item) if method.is_class else partial(self.objc_call, item)

        raise AttributeError(f''''{self.class_.name}' has no attribute {item}''')

    def __getattr__(self, item: str):
        return self[self.class_.sanitize_name(item)]

    def __setitem__(self, key, value):
        if isinstance(key, int):
            super().__setitem__(key, value)
            return

        with suppress(SettingIvarError):
            self._set_ivar(key, value)
            return

    def __setattr__(self, key, value):
        try:
            key = self.__getattribute__('class_').sanitize_name(key)
        except AttributeError:
            pass
        try:
            self._set_ivar(key, value)
        except SettingIvarError:
            super().__setattr__(key, value)

    def __str__(self):
        return self._to_str(False)

    def __repr__(self):
        return f'<{self.__class__.__name__} 0x{int(self):x} Class: {self.class_.name}>'



import pytest

pytestmark = pytest.mark.darwin


def test_modules(client):
    """
    :param rpcclient.client.Client client:
    """
    assert '/usr/lib/libSystem.B.dylib' in [module.name for module in client.images]


def test_uname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.uname.sysname == 'Darwin'
    assert 'machine' in client.uname


def test_get_class_list(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert len(client.get_class_list()) > 0


def test_load_framework(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assert client.load_framework('Foundation') != 0


def test_load_all_libraries(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    original_count = len(client.images)
    client.load_all_libraries()
    assert len(client.images) > original_count


def test_autorelease_pool_object_retain_count(client) -> None:
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    obj = client.cf("tst")
    orig_retain_count = obj.objc_call('retainCount').c_int64
    with client.create_autorelease_pool_ctx():
        obj.objc_call('retain')
        obj.objc_call('autorelease')
        retain_count = obj.objc_call('retainCount').c_int64
        assert retain_count == orig_retain_count + 1
    retain_count = obj.objc_call('retainCount').c_int64
    assert retain_count == orig_retain_count


def test_autorelease_pool_object_retrival(client) -> None:
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.create_autorelease_pool_ctx():
        obj = client.cf('tst')
        pool = client.get_current_autorelease_pool()
        assert obj == pool[0]


def test_autorelease_pool_object_retrival_from_ctx(client) -> None:
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.create_autorelease_pool_ctx() as pool:
        obj = client.cf('tst')
        assert obj == pool.get_autorelease_pool()[0]


def test_autorelease_pool_object_retrival_refresh(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    with client.create_autorelease_pool_ctx() as pool:
        obj = client.cf('tst')
        pool = pool.get_autorelease_pool()
        assert obj in pool
        obj2 = client.cf('tst2')
        assert obj2 not in pool
        pool.refresh()
        assert obj2 in pool



import typing
from contextlib import suppress

from rpcclient.exceptions import BadReturnValueError, NoSuchPreferenceError, RpcClientException

kCFPreferencesCurrentUser = 'kCFPreferencesCurrentUser'
kCFPreferencesAnyUser = 'kCFPreferencesAnyUser'
kCFPreferencesCurrentHost = 'kCFPreferencesCurrentHost'
kCFPreferencesAnyHost = 'kCFPreferencesAnyHost'
GLOBAL_DOMAIN = 'Apple Global Domain'


class CFPreferences:
    """
    API to the CFPreferences* functions - preferences managed by cfprefsd.
    https://developer.apple.com/documentation/corefoundation/preferences_utilities?language=objc
    """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client

    def get_keys(self, application_id: str, username: str = kCFPreferencesCurrentUser,
                 hostname: str = kCFPreferencesCurrentHost) -> typing.Optional[list[str]]:
        """ wrapper for CFPreferencesCopyKeyList """
        application_id = self._client.cf(application_id)
        username = self._client.cf(username)
        hostname = self._client.cf(hostname)
        keys = self._client.symbols.CFPreferencesCopyKeyList(application_id, username, hostname).py()
        if keys is None:
            raise NoSuchPreferenceError()
        return keys

    def get_value(self, key: str, application_id: str, username: str = kCFPreferencesCurrentUser,
                  hostname: str = kCFPreferencesCurrentHost) -> typing.Optional[str]:
        """ wrapper for CFPreferencesCopyValue """
        key = self._client.cf(key)
        application_id = self._client.cf(application_id)
        username = self._client.cf(username)
        hostname = self._client.cf(hostname)
        return self._client.symbols.CFPreferencesCopyValue(key, application_id, username, hostname).py()

    def get_dict(self, application_id: str, username: str = kCFPreferencesCurrentUser,
                 hostname: str = kCFPreferencesCurrentHost) -> typing.Optional[dict]:
        """ get a dictionary representation of given preference """
        result = {}
        key_list = self.get_keys(application_id, username, hostname)
        if not key_list:
            raise RpcClientException(f'failed to get key list for: {application_id}/{username}/{hostname}')
        for k in key_list:
            result[k] = self.get_value(k, application_id, username, hostname)
        return result

    def set(self, key: str, value, application_id: str, username: str = kCFPreferencesCurrentUser,
            hostname: str = kCFPreferencesCurrentHost):
        """ wrapper for CFPreferencesSetValue """
        self._client.symbols.CFPreferencesSetValue(self._client.cf(key), self._client.cf(value),
                                                   self._client.cf(application_id), self._client.cf(username),
                                                   self._client.cf(hostname))

    def remove(self, key: str, application_id: str, username: str = kCFPreferencesCurrentUser,
               hostname: str = kCFPreferencesCurrentHost):
        """ remove a given key from a preference """
        self._client.symbols.CFPreferencesSetValue(self._client.cf(key), 0,
                                                   self._client.cf(application_id), self._client.cf(username),
                                                   self._client.cf(hostname))

    def set_dict(self, d: dict, application_id: str, username: str = kCFPreferencesCurrentUser,
                 hostname: str = kCFPreferencesCurrentHost):
        """ set entire preference dictionary (erase first if exists) """
        with suppress(NoSuchPreferenceError):
            self.clear(application_id, username, hostname)
        self.update_dict(d, application_id, username, hostname)

    def update_dict(self, d: dict, application_id: str, username: str = kCFPreferencesCurrentUser,
                    hostname: str = kCFPreferencesCurrentHost):
        """ update preference dictionary """
        for k, v in d.items():
            self.set(k, v, application_id, username, hostname)

    def clear(self, application_id: str, username: str = kCFPreferencesCurrentUser,
              hostname: str = kCFPreferencesCurrentHost):
        """ remove all values from given preference """
        for k in self.get_keys(application_id, username, hostname):
            self.remove(k, application_id, username, hostname)

    def sync(self, application_id: str, username: str = kCFPreferencesCurrentUser,
             hostname: str = kCFPreferencesCurrentHost):
        if not self._client.symbols.CFPreferencesSynchronize(self._client.cf(application_id), self._client.cf(username),
                                                             self._client.cf(hostname)):
            raise BadReturnValueError('CFPreferencesSynchronize() failed')



from collections import UserList
from typing import List, Optional

from construct import Hex, Int32ul, PaddedString, Struct

from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.structs.generic import SymbolFormatField

magic_t = Struct(
    'm0' / Hex(Int32ul),
    'm1' / PaddedString(12, 'ascii')
)


SIZEOF_PAGE_DATA = 0x38


def AutoreleasePoolPageData(client) -> Struct:
    return Struct(
            'magic' / magic_t,
            'next' / SymbolFormatField(client),
            'thread' / SymbolFormatField(client),
            'parent' / SymbolFormatField(client),
            'child' / SymbolFormatField(client),
            'depth' / Hex(Int32ul),
            'hiwat' / Hex(Int32ul)
        )


class AutoreleasePool(UserList[DarwinSymbol]):
    """
    A list-like container for `DarwinSymbol` objects representing one Objective-C autorelease pool.
    """

    def __init__(self, client, address: DarwinSymbol) -> None:
        """
        Initialize `AutoreleasePool`.

        :param client:
        :param address: address of the pool start.
        """
        super().__init__()
        self._client = client
        self._mask = client.symbols.objc_debug_autoreleasepoolpage_ptr_mask[0]
        self.address = address
        self.end = address + 8
        self.refresh()

    def refresh(self) -> None:
        """
        refresh the content of of the `AutoreleasePool` object to reflect the current state of the actual pool
        """
        self.clear()
        get_autorelease_pool_end(self._client)  # to solve autorelease pool bug
        page_sym = find_page_for_address(self._client, self.address)
        page = AutoreleasePoolPageData(self._client).parse_stream(page_sym)
        next = self.address + 8
        while (next[0].c_int64 not in [0, 0xa3a3a3a3, 0xa3a3a3a3a3a3a3a3] or
               next == page.next):
            if next == page.next:
                if page.child == 0:
                    break
                next = page.child + SIZEOF_PAGE_DATA
                page = AutoreleasePoolPageData(self._client).parse_stream(
                    page.child)
                continue
            self.append(next[0] & self._mask)
            next += 8
        self.end = next

    def __repr__(self) -> str:
        """
        Return human-readable representation showing pool name and object count.

        :return: String representation of the pool
        """
        return f'<{self.__class__.__name__} {hex(self.address)} contains {len(self)} objects>'

    def __str__(self) -> str:
        """
        Return string representation showing pool name all of the objects inside.

        :return: String representation of the pool
        """
        output = f'{self.__class__.__name__} {hex(self.address)}:\n'
        for idx in range(len(self)):
            class_name = self._client.symbols.class_getName(self[idx].objc_call('class')).peek_str()
            output += f'#{idx}:\t0x{self[idx]:x}\t{class_name}\n'

        return output


class AutorelesePoolCtx:
    """
    Context manager to create and drain an Objective-C `NSAutoreleasePool`.

    On enter: creates a new pool.
    On exit: drains (releases) the pool.
    """

    def __init__(self, client) -> None:
        """
        Initialize `AutorelesePoolCtx`.

        :param client:
        """
        self._client = client
        self._pool = None
        self._create()

    def __enter__(self) -> 'AutorelesePoolCtx':
        """
        Ensure a fresh pool exists when entering `with` block.

        :return: Self reference for the context manager
        """
        self._create()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """
        Drain the pool when exiting `with` block.
        """
        self.drain()

    def _create(self) -> None:
        """
        Create the pool if not already present.
        """
        if self._pool is None:
            self._pool = self._client.symbols.objc_autoreleasePoolPush()

    def drain(self) -> None:
        """
        Drain (release) the current autorelease pool if it exists, then clear it.
        """
        if self._pool is not None:
            self._client.symbols.objc_autoreleasePoolPop(self._pool)
            self._pool = None

    def get_autorelease_pool(self) -> Optional[AutoreleasePool]:
        """
        Return an `AutoreleasePool` object representing the current pool.

        :return: `AutoreleasePool` instance if pool is initialized, else None
        """
        if self._pool is not None:
            return AutoreleasePool(self._client, self._pool)
        return None


def get_autorelease_pool_end(client) -> DarwinSymbol:
    """
    Retreive the end of the autorelease pool.

    :param client:
    :return: `DarwinSymbol` representing the end of the pool
    """
    end = client.symbols.objc_autoreleasePoolPush()
    client.symbols.objc_autoreleasePoolPop(end)
    return end


def find_page_for_address(client, address: DarwinSymbol) -> DarwinSymbol:
    """
    Find the page for a given address inside of the thread's autorelease pool.

    :param client:
    :param address: An address inside of the autorelease pool
    :return: `DarwinSymbol` representing page
    """
    current = address - SIZEOF_PAGE_DATA
    while current.peek(4) != b'\xa1\xa1\xa1\xa1':
        current -= 8
    return current


def find_hot_page(client) -> DarwinSymbol:
    """
    Find the current hot page of the thread's autorelease pool.

    :param client:
    :return: `DarwinSymbol` representing the end of the pool
    """
    return find_page_for_address(client, get_autorelease_pool_end(client))


def find_first_page(client) -> DarwinSymbol:
    """
    Find the current first page of the thread's autorelease pool.

    :param client:
    :return: `DarwinSymbol` representing the end of the pool
    """
    page_sym = find_hot_page(client)
    page = AutoreleasePoolPageData(client).parse_stream(page_sym)
    while page.parent != 0:
        page_sym = page.parent
        page = AutoreleasePoolPageData(client).parse_stream(page_sym)
    return page_sym


def get_autorelease_pools(client) -> List[AutoreleasePool]:
    """
    Get and parse all autorelease pools currently in the thread.

    :param client:
    :return: List of `AutoreleasePool` instances found in the dump
    """
    end = get_autorelease_pool_end(client)
    page_sym = find_first_page(client)
    pool = AutoreleasePool(client, page_sym + SIZEOF_PAGE_DATA)
    pools: List[AutoreleasePool] = [pool]
    while pool.end != end:
        pool = AutoreleasePool(client, pool.end)
        pools.append(pool)
    return pools


def get_current_autorelease_pool(client) -> AutoreleasePool:
    """
    Get the most recently created autorelease pool.

    :param client:
    :return: The last `AutoreleasePool` in the list (most recent)
    :raises IndexError: if no pools are found
    """
    return get_autorelease_pools(client)[-1]



import logging
from datetime import datetime, timedelta
from typing import Optional

from rpcclient.darwin.processes import DarwinProcesses, Process
from rpcclient.exceptions import LaunchError
from rpcclient.structs.consts import SIGKILL

logger = logging.getLogger(__name__)


class IosProcesses(DarwinProcesses):
    def launch(self, bundle_id: str, kill_existing: bool = True, timeout: float = 1, unlock_device: bool = True,
               disable_aslr: bool = False, wait_for_debugger: bool = False, stdout: Optional[str] = None,
               stderr: Optional[str] = None) -> Process:
        """
        launch process using BackBoardService
        https://github.com/swigger/debugserver-ios/blob/master/inc/BackBoardServices.framework/Headers/BackBoardServices.h
        """
        debug_options = {}
        options = {}
        sym = self._client.symbols
        debug_options[sym.BKSDebugOptionKeyDisableASLR[0].py()] = disable_aslr
        debug_options[sym.BKSDebugOptionKeyWaitForDebugger[0].py()] = wait_for_debugger
        if stdout is not None:
            debug_options[sym.BKSDebugOptionKeyStandardOutPath[0].py()] = stdout
        if stderr is not None:
            debug_options[sym.BKSDebugOptionKeyStandardErrorPath[0].py()] = stderr
        options[sym.BKSOpenApplicationOptionKeyUnlockDevice[0].py()] = unlock_device
        options[sym.BKSOpenApplicationOptionKeyDebuggingOptions[0].py()] = debug_options

        bkssystem_service = self._client.symbols.objc_getClass('BKSSystemService').objc_call('new')
        pid = bkssystem_service.objc_call('pidForApplication:', self._client.cf(bundle_id)).c_int32
        if pid != -1 and kill_existing:
            logger.info(f'Kill existing process {pid}')
            self.kill(pid, SIGKILL)

        bkssystem_service.objc_call(
            'openApplication:options:clientPort:withResult:', self._client.cf(bundle_id), self._client.cf(options),
            bkssystem_service.objc_call('createClientPort'), self._client.get_dummy_block())

        start_time = datetime.now()
        timeout = timedelta(seconds=timeout)
        while datetime.now() - start_time < timeout:
            pid = bkssystem_service.objc_call('pidForApplication:', self._client.cf(bundle_id)).c_int32

        if pid == -1:
            raise LaunchError()
        return self.get_by_pid(pid)



import logging
import os
import subprocess
import typing
import urllib.request
from socket import socket

import requests

from rpcclient.client import Client
from rpcclient.exceptions import FailedToConnectError
from rpcclient.ios.client import IosClient
from rpcclient.linux.client import LinuxClient
from rpcclient.macos.client import MacosClient
from rpcclient.protosocket import ProtoSocket

PROJECT_URL = 'https://api.github.com/repos/doronz88/rpc-project/releases/latest'
BINARY_NAME = 'rpcserver_macosx'

logger = logging.getLogger(__name__)

DEFAULT_PORT = 5910


def create_local() -> typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    response = requests.get(PROJECT_URL)
    response.raise_for_status()
    data = response.json()
    for asset in data['assets']:
        if asset['name'] == BINARY_NAME:
            download_url = asset['browser_download_url']
            logger.info(f'Downloading {BINARY_NAME} from {download_url}')
            urllib.request.urlretrieve(download_url, BINARY_NAME)
            logger.info(f'Download completed: {BINARY_NAME}')
            os.chmod(BINARY_NAME, os.stat(BINARY_NAME).st_mode | 0o100)

    server = socket()
    server.bind(('127.0.0.1', 0))
    available_port = server.getsockname()[1]
    server.close()

    proc = subprocess.Popen([f'./{BINARY_NAME}', '-p', str(available_port)], cwd=os.getcwd())
    logger.info(f'server launched at pid {proc.pid}')

    while True:
        try:
            return create_tcp_client('127.0.0.1', available_port)
        except FailedToConnectError:
            pass


def create_tcp_client(hostname: str, port: int = DEFAULT_PORT) -> \
        typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    def tcp_connect() -> socket:
        sock = socket()
        try:
            sock.connect((hostname, port))
        except ConnectionRefusedError as e:
            # wrap in our own exception
            raise FailedToConnectError() from e
        return sock

    return create_client(tcp_connect)


def create_client(create_socket_cb: typing.Callable) -> typing.Union[Client, IosClient, MacosClient, LinuxClient]:
    sock = create_socket_cb()
    proto_sock = ProtoSocket(sock)

    sysname = proto_sock.handshake.sysname.lower()
    machine = proto_sock.handshake.machine.lower()
    arch = proto_sock.handshake.arch

    logging.info(f'connection uname.sysname: {sysname} uname.machine: {machine}')

    if sysname == 'darwin':
        if machine.startswith('iphone'):
            return IosClient(proto_sock, sysname, arch, create_socket_cb)
        else:
            return MacosClient(proto_sock, sysname, arch, create_socket_cb)
    elif sysname == 'linux':
        return LinuxClient(proto_sock, sysname, arch, create_socket_cb)

    return Client(proto_sock, sysname, arch, create_socket_cb)



from pathlib import Path

from rpcclient.darwin.crash_reports import CrashReports


class Reports:
    """ equivalent to the data that can be found using the Console app inside the Reports section """

    def __init__(self, client, crash_reports_dir):
        self._client = client
        self.crash_reports = CrashReports(client, crash_reports_dir)

    def get_logs(self, prefix='') -> list[Path]:
        result = []
        sub_paths = ['var/log', 'Library/Logs']
        for sub_path in sub_paths:
            for path in self._client.roots:
                path = Path(path) / sub_path
                if not self._client.fs.accessible(path):
                    continue

                for root, dirs, files in self._client.fs.walk(path, onerror=lambda x: None):
                    for name in files:
                        if not self._client.fs.accessible(path):
                            continue

                        if name.endswith('.log') and name.startswith(prefix):
                            result.append(Path(root) / name)
        return result

    @property
    def system_log(self) -> str:
        with self._client.fs.open('/var/log/system.log', 'r') as f:
            return f.read().decode()



from pathlib import Path

from pycrashreport.crash_report import CrashReportBase, get_crash_report_from_buf


class CrashReports:
    """" manage crash reports """

    def __init__(self, client, crash_reports_dir):
        self._client = client
        self._crash_reports_dir = crash_reports_dir

    def set_symbolicated(self, enabled: bool = True):
        """
        enable/disable crash reports symbolication
        https://github.com/dlevi309/Symbolicator
        """
        self._client.preferences.cf.set('SymbolicateCrashes', enabled, 'com.apple.CrashReporter', 'root')

        # bugfix: at some point, this setting was moved to "com.apple.osanalytics" bundle identifier
        self._client.preferences.cf.set('SymbolicateCrashes', enabled, 'com.apple.osanalytics', 'root')

    def list(self, prefixed='') -> list[CrashReportBase]:
        """ get a list of all crash reports as CrashReport parsed objects """
        result = []
        for root in self._client.roots:
            root = Path(root) / self._crash_reports_dir

            if not self._client.fs.accessible(root):
                continue

            for entry in self._client.fs.scandir(root):
                if entry.is_file() and entry.name.endswith('.ips') and entry.name.startswith(prefixed):
                    with self._client.fs.open(entry.path, 'r') as f:
                        result.append(get_crash_report_from_buf(f.read().decode(), filename=entry.path))
        return result

    def clear(self, prefixed=''):
        """ remove all existing crash reports """
        for entry in self.list(prefixed=prefixed):
            self._client.fs.remove(entry.filename)



import pytest

pytestmark = pytest.mark.darwin


def test_from_xpc_object(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    criteria = client.symbols.xpc_dictionary_create(0, 0, 0)
    client.symbols.xpc_dictionary_set_int64(criteria, 'Delay', 5)
    client.symbols.xpc_dictionary_set_int64(criteria, 'GracePeriod', 1)
    client.symbols.xpc_dictionary_set_string(criteria, 'Priority', 'Utility')
    assert client.xpc.decode_xpc_object_using_cf_serialization(criteria) == {
        'Delay': 5, 'Priority': 'Utility', 'GracePeriod': 1}



class MobileGestalt:
    def __init__(self, client):
        self._client = client

    # Identifying Information

    @property
    def DiskUsage(self):
        return self.get_answer('DiskUsage')

    @property
    def ModelNumber(self):
        return self.get_answer('ModelNumber')

    @property
    def SIMTrayStatus(self):
        return self.get_answer('SIMTrayStatus')

    @property
    def SerialNumber(self):
        return self.get_answer('SerialNumber')

    @property
    def MLBSerialNumber(self):
        return self.get_answer('MLBSerialNumber')

    @property
    def UniqueDeviceID(self):
        return self.get_answer('UniqueDeviceID')

    @property
    def UniqueDeviceIDData(self):
        return self.get_answer('UniqueDeviceIDData')

    @property
    def UniqueChipID(self):
        return self.get_answer('UniqueChipID')

    @property
    def InverseDeviceID(self):
        return self.get_answer('InverseDeviceID')

    @property
    def DiagData(self):
        return self.get_answer('DiagData')

    @property
    def DieId(self):
        return self.get_answer('DieId')

    @property
    def CPUArchitecture(self):
        return self.get_answer('CPUArchitecture')

    @property
    def PartitionType(self):
        return self.get_answer('PartitionType')

    @property
    def UserAssignedDeviceName(self):
        return self.get_answer('UserAssignedDeviceName')

    # Bluetooth Information

    @property
    def BluetoothAddress(self):
        return self.get_answer('BluetoothAddress')

    # Battery Information

    @property
    def RequiredBatteryLevelForSoftwareUpdate(self):
        return self.get_answer('RequiredBatteryLevelForSoftwareUpdate')

    @property
    def BatteryIsFullyCharged(self):
        return self.get_answer('BatteryIsFullyCharged')

    @property
    def BatteryIsCharging(self):
        return self.get_answer('BatteryIsCharging')

    @property
    def BatteryCurrentCapacity(self):
        return self.get_answer('BatteryCurrentCapacity')

    @property
    def ExternalPowerSourceConnected(self):
        return self.get_answer('ExternalPowerSourceConnected')

    # Baseband Information

    @property
    def BasebandSerialNumber(self):
        return self.get_answer('BasebandSerialNumber')

    @property
    def BasebandCertId(self):
        return self.get_answer('BasebandCertId')

    @property
    def BasebandChipId(self):
        return self.get_answer('BasebandChipId')

    @property
    def BasebandFirmwareManifestData(self):
        return self.get_answer('BasebandFirmwareManifestData')

    @property
    def BasebandFirmwareVersion(self):
        return self.get_answer('BasebandFirmwareVersion')

    @property
    def BasebandKeyHashInformation(self):
        return self.get_answer('BasebandKeyHashInformation')

    # Telephony Information

    @property
    def CarrierBundleInfoArray(self):
        return self.get_answer('CarrierBundleInfoArray')

    @property
    def CarrierInstallCapability(self):
        return self.get_answer('CarrierInstallCapability')

    @property
    def InternationalMobileEquipmentIdentity(self):
        return self.get_answer('InternationalMobileEquipmentIdentity')

    @property
    def MobileSubscriberCountryCode(self):
        return self.get_answer('MobileSubscriberCountryCode')

    @property
    def MobileSubscriberNetworkCode(self):
        return self.get_answer('MobileSubscriberNetworkCode')

    # Device Information

    @property
    def ChipID(self):
        return self.get_answer('ChipID')

    @property
    def ComputerName(self):
        return self.get_answer('ComputerName')

    @property
    def DeviceVariant(self):
        return self.get_answer('DeviceVariant')

    @property
    def HWModelStr(self):
        return self.get_answer('HWModelStr')

    @property
    def BoardId(self):
        return self.get_answer('BoardId')

    @property
    def HardwarePlatform(self):
        return self.get_answer('HardwarePlatform')

    @property
    def DeviceName(self):
        return self.get_answer('DeviceName')

    @property
    def DeviceColor(self):
        return self.get_answer('DeviceColor')

    @property
    def DeviceClassNumber(self):
        return self.get_answer('DeviceClassNumber')

    @property
    def DeviceClass(self):
        return self.get_answer('DeviceClass')

    @property
    def BuildVersion(self):
        return self.get_answer('BuildVersion')

    @property
    def ProductName(self):
        return self.get_answer('ProductName')

    @property
    def ProductType(self):
        return self.get_answer('ProductType')

    @property
    def ProductVersion(self):
        return self.get_answer('ProductVersion')

    @property
    def FirmwareNonce(self):
        return self.get_answer('FirmwareNonce')

    @property
    def FirmwareVersion(self):
        return self.get_answer('FirmwareVersion')

    @property
    def FirmwarePreflightInfo(self):
        return self.get_answer('FirmwarePreflightInfo')

    @property
    def IntegratedCircuitCardIdentifier(self):
        return self.get_answer('IntegratedCircuitCardIdentifier')

    @property
    def AirplaneMode(self) -> bool:
        return self.get_answer('AirplaneMode')

    @property
    def AllowYouTube(self):
        return self.get_answer('AllowYouTube')

    @property
    def AllowYouTubePlugin(self):
        return self.get_answer('AllowYouTubePlugin')

    @property
    def MinimumSupportediTunesVersion(self):
        return self.get_answer('MinimumSupportediTunesVersion')

    @property
    def ProximitySensorCalibration(self):
        return self.get_answer('ProximitySensorCalibration')

    @property
    def RegionCode(self):
        return self.get_answer('RegionCode')

    @property
    def RegionInfo(self):
        return self.get_answer('RegionInfo')

    @property
    def RegulatoryIdentifiers(self):
        return self.get_answer('RegulatoryIdentifiers')

    @property
    def SBAllowSensitiveUI(self):
        return self.get_answer('SBAllowSensitiveUI')

    @property
    def SBCanForceDebuggingInfo(self):
        return self.get_answer('SBCanForceDebuggingInfo')

    @property
    def SDIOManufacturerTuple(self):
        return self.get_answer('SDIOManufacturerTuple')

    @property
    def SDIOProductInfo(self):
        return self.get_answer('SDIOProductInfo')

    @property
    def ShouldHactivate(self):
        return self.get_answer('ShouldHactivate')

    @property
    def SigningFuse(self):
        return self.get_answer('SigningFuse')

    @property
    def SoftwareBehavior(self):
        return self.get_answer('SoftwareBehavior')

    @property
    def SoftwareBundleVersion(self):
        return self.get_answer('SoftwareBundleVersion')

    @property
    def SupportedDeviceFamilies(self):
        return self.get_answer('SupportedDeviceFamilies')

    @property
    def SupportedKeyboards(self):
        return self.get_answer('SupportedKeyboards')

    @property
    def TotalSystemAvailable(self):
        return self.get_answer('TotalSystemAvailable')

    # Capability Information

    @property
    def AllDeviceCapabilities(self):
        return self.get_answer('AllDeviceCapabilities')

    @property
    def AppleInternalInstallCapability(self):
        return self.get_answer('AppleInternalInstallCapability')

    @property
    def ExternalChargeCapability(self):
        return self.get_answer('ExternalChargeCapability')

    @property
    def ForwardCameraCapability(self):
        return self.get_answer('ForwardCameraCapability')

    @property
    def PanoramaCameraCapability(self):
        return self.get_answer('PanoramaCameraCapability')

    @property
    def RearCameraCapability(self):
        return self.get_answer('RearCameraCapability')

    @property
    def HasAllFeaturesCapability(self):
        return self.get_answer('HasAllFeaturesCapability')

    @property
    def HasBaseband(self):
        return self.get_answer('HasBaseband')

    @property
    def HasInternalSettingsBundle(self):
        return self.get_answer('HasInternalSettingsBundle')

    @property
    def HasSpringBoard(self):
        return self.get_answer('HasSpringBoard')

    @property
    def InternalBuild(self):
        return self.get_answer('InternalBuild')

    @property
    def IsSimulator(self):
        return self.get_answer('IsSimulator')

    @property
    def IsThereEnoughBatteryLevelForSoftwareUpdate(self):
        return self.get_answer('IsThereEnoughBatteryLevelForSoftwareUpdate')

    @property
    def IsUIBuild(self):
        return self.get_answer('IsUIBuild')

    @property
    def PasswordConfigured(self):
        return self.get_answer('PasswordConfigured')

    @property
    def PasswordProtected(self):
        return self.get_answer('PasswordProtected')

    # Regional Behaviour

    @property
    def RegionalBehaviorAll(self):
        return self.get_answer('RegionalBehaviorAll')

    @property
    def RegionalBehaviorChinaBrick(self):
        return self.get_answer('RegionalBehaviorChinaBrick')

    @property
    def RegionalBehaviorEUVolumeLimit(self):
        return self.get_answer('RegionalBehaviorEUVolumeLimit')

    @property
    def RegionalBehaviorGB18030(self):
        return self.get_answer('RegionalBehaviorGB18030')

    @property
    def RegionalBehaviorGoogleMail(self):
        return self.get_answer('RegionalBehaviorGoogleMail')

    @property
    def RegionalBehaviorNTSC(self):
        return self.get_answer('RegionalBehaviorNTSC')

    @property
    def RegionalBehaviorNoPasscodeLocationTiles(self):
        return self.get_answer('RegionalBehaviorNoPasscodeLocationTiles')

    @property
    def RegionalBehaviorNoVOIP(self):
        return self.get_answer('RegionalBehaviorNoVOIP')

    @property
    def RegionalBehaviorNoWiFi(self):
        return self.get_answer('RegionalBehaviorNoWiFi')

    @property
    def RegionalBehaviorShutterClick(self):
        return self.get_answer('RegionalBehaviorShutterClick')

    @property
    def RegionalBehaviorVolumeLimit(self):
        return self.get_answer('RegionalBehaviorVolumeLimit')

    # Wireless Information

    @property
    def ActiveWirelessTechnology(self):
        return self.get_answer('ActiveWirelessTechnology')

    @property
    def WifiAddress(self):
        return self.get_answer('WifiAddress')

    @property
    def WifiAddressData(self):
        return self.get_answer('WifiAddressData')

    @property
    def WifiVendor(self):
        return self.get_answer('WifiVendor')

    # FaceTime Information

    @property
    def FaceTimeBitRate2G(self):
        return self.get_answer('FaceTimeBitRate2G')

    @property
    def FaceTimeBitRate3G(self):
        return self.get_answer('FaceTimeBitRate3G')

    @property
    def FaceTimeBitRateLTE(self):
        return self.get_answer('FaceTimeBitRateLTE')

    @property
    def FaceTimeBitRateWiFi(self):
        return self.get_answer('FaceTimeBitRateWiFi')

    @property
    def FaceTimeDecodings(self):
        return self.get_answer('FaceTimeDecodings')

    @property
    def FaceTimeEncodings(self):
        return self.get_answer('FaceTimeEncodings')

    @property
    def FaceTimePreferredDecoding(self):
        return self.get_answer('FaceTimePreferredDecoding')

    @property
    def FaceTimePreferredEncoding(self):
        return self.get_answer('FaceTimePreferredEncoding')

    # More Device Capabilities

    @property
    def DeviceSupportsFaceTime(self):
        return self.get_answer('DeviceSupportsFaceTime')

    @property
    def DeviceSupportsTethering(self):
        return self.get_answer('DeviceSupportsTethering')

    @property
    def DeviceSupportsSimplisticRoadMesh(self):
        return self.get_answer('DeviceSupportsSimplisticRoadMesh')

    @property
    def DeviceSupportsNavigation(self):
        return self.get_answer('DeviceSupportsNavigation')

    @property
    def DeviceSupportsLineIn(self):
        return self.get_answer('DeviceSupportsLineIn')

    @property
    def DeviceSupports9Pin(self):
        return self.get_answer('DeviceSupports9Pin')

    @property
    def DeviceSupports720p(self):
        return self.get_answer('DeviceSupports720p')

    @property
    def DeviceSupports4G(self):
        return self.get_answer('DeviceSupports4G')

    @property
    def DeviceSupports3DMaps(self):
        return self.get_answer('DeviceSupports3DMaps')

    @property
    def DeviceSupports3DImagery(self):
        return self.get_answer('DeviceSupports3DImagery')

    @property
    def DeviceSupports1080p(self):
        return self.get_answer('DeviceSupports1080p')

    def get_answer(self, key: str):
        """ get a string answer from MobileGestalt """
        return self._client.symbols.MGCopyAnswer(self._client.cf(key)).py()

    def set_answer(self, key: str, value):
        """ set an answer into MobileGestalt """
        return self._client.symbols.MGSetAnswer(self._client.cf(key), self._client.cf(value))



import ast
import builtins
import contextlib
import ctypes
import dataclasses
import enum
import logging
import os
import sys
import threading
import typing
from collections import namedtuple
from pathlib import Path
from select import select
from typing import Any, Optional

import IPython
from traitlets.config import Config
from xonsh.built_ins import XSH
from xonsh.main import main as xonsh_main

import rpcclient
from rpcclient.capture_fd import CaptureFD
from rpcclient.exceptions import ArgumentError, BadReturnValueError, RpcBrokenPipeError, RpcConnectionRefusedError, \
    RpcFileExistsError, RpcFileNotFoundError, RpcIsADirectoryError, RpcNotADirectoryError, RpcNotEmptyError, \
    RpcPermissionError, RpcResourceTemporarilyUnavailableError, ServerResponseError, SpawnError, SymbolAbsentError
from rpcclient.fs import Fs
from rpcclient.lief import Lief
from rpcclient.network import Network
from rpcclient.processes import Processes
from rpcclient.protobuf_bridge import Argument, CmdCall, CmdDlclose, CmdDlopen, CmdDlsym, CmdDummyBlock, CmdExec, \
    CmdListDir, CmdPeek, CmdPoke, Response
from rpcclient.protosocket import ProtoSocket
from rpcclient.structs.consts import EAGAIN, ECONNREFUSED, EEXIST, EISDIR, ENOENT, ENOTDIR, ENOTEMPTY, EPERM, EPIPE, \
    RTLD_NEXT
from rpcclient.symbol import Symbol
from rpcclient.symbols_jar import SymbolsJar
from rpcclient.sysctl import Sysctl

tty_support = False
try:
    import termios
    import tty

    tty_support = True
except ImportError:
    logging.warning('termios not available on your system. some functionality may not work as expected')

io_or_str = typing.TypeVar('io_or_str', typing.IO, str)

SpawnResult = namedtuple('SpawnResult', 'error pid stdout')

INVALID_PID = 0xffffffff
CHUNK_SIZE = 1024

USAGE = '''
Welcome to the rpcclient interactive shell! You interactive shell for controlling the remote rpcserver.
Feel free to use the following globals:

🌍 p - the injected process
🌍 symbols - process global symbols

Have a nice flight ✈️!
Starting an IPython shell... 🐍
'''


@dataclasses.dataclass
class ProtocolDitentStat:
    errno: int
    st_dev: int  # device inode resides on
    st_mode: int  # inode protection mode
    st_nlink: int  # number of hard links to the file
    st_ino: int  # inode's number
    st_uid: int  # user-id of owner
    st_gid: int  # group-id of owner
    st_rdev: int  # device type, for special file inode
    st_size: int  # file size, in bytes
    st_blocks: int  # blocks allocated for file
    st_blksize: int  # optimal blocksize for I/O
    st_atime: int
    st_mtime: int
    st_ctime: int


@dataclasses.dataclass
class ProtocolDirent:
    d_inode: int
    d_type: int
    d_name: str
    lstat: ProtocolDitentStat
    stat: ProtocolDitentStat


class Client:
    """ Main client interface to access remote rpcserver """

    DEFAULT_ARGV = ['/bin/sh']
    DEFAULT_ENVP = []

    def __init__(self, sock: ProtoSocket, sysname: str, arch, create_socket_cb: typing.Callable,
                 dlsym_global_handle=RTLD_NEXT):
        self._arch = arch
        self._create_socket_cb = create_socket_cb
        self._sock = sock
        self._old_settings = None
        self._endianness = '<'
        self._sysname = sysname
        self._dlsym_global_handle = dlsym_global_handle
        self._protocol_lock = threading.Lock()
        self._logger = logging.getLogger(self.__module__)
        self._ipython_run_cell_hook_enabled = True

        self.reconnect_lock = threading.Lock()

        self._init_process_specific()

    def _init_process_specific(self):
        self.symbols = SymbolsJar.create(self)
        self.fs = Fs(self)
        self.processes = Processes(self)
        self.network = Network(self)
        self.lief = Lief(self)
        self.sysctl = Sysctl(self)

    def info(self):
        """ print information about current target """
        uname = self.uname
        print('sysname:', uname.sysname)
        print('nodename:', uname.nodename)
        print('release:', uname.release)
        print('version:', uname.version)
        print('machine:', uname.machine)
        print(f'uid: {self.symbols.getuid():d}')
        print(f'gid: {self.symbols.getgid():d}')
        print(f'pid: {self.symbols.getpid():d}')
        print(f'ppid: {self.symbols.getppid():d}')
        print(f'progname: {self.symbols.getprogname().peek_str()}')

    @property
    def uname(self):
        """ get the utsname struct from remote """
        raise NotImplementedError()

    @property
    def arch(self):
        """ get remote arch """
        return self._arch

    def dlopen(self, filename: str, mode: int) -> Symbol:
        """ call dlopen() at remote and return its handle. see the man page for more details. """
        command = CmdDlopen(filename=filename, mode=mode)
        response = self._sock.send_recv(command)
        return self.symbol(response.handle)

    def dlclose(self, lib: int):
        """ call dlclose() at remote and return its handle. see the man page for more details. """
        command = CmdDlclose(handle=ctypes.c_uint64(lib).value)
        response = self._sock.send_recv(command)
        return response.res

    def dlsym(self, lib: int, symbol_name: str):
        """ call dlsym() at remote and return its handle. see the man page for more details. """
        command = CmdDlsym(handle=ctypes.c_uint64(lib).value, symbol_name=symbol_name)
        response = self._sock.send_recv(command)
        return response.ptr

    def call(self, address: int, argv: list[int] = None, return_float64=False, return_float32=False,
             return_raw=False, va_list_index: int = 0xffff) -> typing.Union[float, Symbol, Any]:
        """ call a remote function and retrieve its return value as Symbol object """
        args = []
        for arg in argv:
            if isinstance(arg, float):
                args.append(Argument(v_double=arg))
            elif isinstance(arg, str):
                args.append(Argument(v_str=arg))
            elif isinstance(arg, int):
                args.append(Argument(v_int=ctypes.c_uint64(arg).value))
            elif isinstance(arg, bytes):
                args.append(Argument(v_bytes=arg))
            elif isinstance(arg, enum.Enum):
                args.append(Argument(v_int=ctypes.c_uint64(arg.value).value))
            else:
                raise ArgumentError()

        command = CmdCall(address=address, va_list_index=va_list_index, argv=args)
        response = self._sock.send_recv(command)
        if response.HasField('arm_registers'):
            double = response.arm_registers.d0
            if return_float32:
                return ctypes.c_float(double).value
            if return_float64:
                return double
            if return_raw:
                return response.arm_registers
            return self.symbol(response.arm_registers.x0)
        return self.symbol(response.return_value)

    def peek(self, address: int, size: int) -> bytes:
        """ peek data at given address """
        command = CmdPeek(address=address, size=size)
        try:
            return self._sock.send_recv(command).data
        except ServerResponseError:
            raise ArgumentError()

    def poke(self, address: int, data: bytes):
        """ poke data at given address """
        command = CmdPoke(address=address, data=data)
        try:
            self._sock.send_recv(command)
        except ServerResponseError:
            raise ArgumentError()

    def get_dummy_block(self) -> Symbol:
        """ get an address for a stub block containing nothing """
        command = CmdDummyBlock()
        response = self._sock.send_recv(command)
        return self.symbol(response.address)

    def listdir(self, filename: str):
        """ get an address for a stub block containing nothing """
        command = CmdListDir(path=filename)
        entries = []
        try:
            response = self._sock.send_recv(command)
        except ServerResponseError:
            self.raise_errno_exception(f'failed to listdir: {filename}')

        for entry in response.dir_entries:
            lstat = ProtocolDitentStat(
                errno=entry.lstat.errno1, st_blocks=entry.lstat.st_blocks, st_blksize=entry.lstat.st_blksize,
                st_atime=entry.lstat.st_atime1, st_ctime=entry.lstat.st_ctime1, st_mtime=entry.lstat.st_mtime1,
                st_nlink=entry.lstat.st_nlink, st_mode=entry.lstat.st_mode, st_rdev=entry.lstat.st_rdev,
                st_size=entry.lstat.st_size, st_dev=entry.lstat.st_dev, st_gid=entry.lstat.st_gid,
                st_ino=entry.lstat.st_ino, st_uid=entry.lstat.st_uid)
            stat = ProtocolDitentStat(
                errno=entry.stat.errno1, st_blocks=entry.stat.st_blocks, st_blksize=entry.stat.st_blksize,
                st_atime=entry.stat.st_atime1, st_ctime=entry.stat.st_ctime1, st_mtime=entry.stat.st_mtime1,
                st_nlink=entry.stat.st_nlink, st_mode=entry.stat.st_mode, st_rdev=entry.stat.st_rdev,
                st_size=entry.stat.st_size, st_dev=entry.stat.st_dev, st_gid=entry.stat.st_gid,
                st_ino=entry.stat.st_ino, st_uid=entry.stat.st_uid)
            entries.append(
                ProtocolDirent(d_inode=entry.lstat.st_ino, d_type=entry.d_type, d_name=entry.d_name,
                               lstat=lstat,
                               stat=stat))
        return entries

    def spawn(self, argv: list[str] = None, envp: list[str] = None, stdin: io_or_str = sys.stdin,
              stdout=sys.stdout, raw_tty=False, background=False) -> SpawnResult:
        """
        spawn a new process and forward its stdin, stdout & stderr

        :param argv: argv of the process to be executed
        :param envp: envp of the process to be executed
        :param stdin: either a file object to read from OR a string
        :param stdout: a file object to write both stdout and stderr to. None if background is requested
        :param raw_tty: should enable raw tty mode
        :param background: should execute process in background
        :return: a SpawnResult. error is None if background is requested
        """
        if argv is None:
            argv = self.DEFAULT_ARGV

        if envp is None:
            envp = self.DEFAULT_ENVP

        with self._protocol_lock:
            try:
                pid = self._execute(argv, envp, background=background)
            except SpawnError:
                # depends on where the error occurred, the socket might be closed
                raise

            self._logger.info(f'shell process started as pid: {pid}')

            if background:
                return SpawnResult(error=None, pid=pid, stdout=None)

            if raw_tty:
                self._prepare_terminal()
            try:
                # the socket must be non-blocking for using select()
                self._sock.raw_socket.setblocking(False)
                error = self._execution_loop(stdin, stdout)
            except Exception:  # noqa: E722
                # this is important to really catch every exception here, even exceptions not inheriting from Exception
                # so the controlling terminal will remain working with its previous settings
                if raw_tty:
                    self._restore_terminal()
                raise
            finally:
                self._sock.raw_socket.setblocking(True)

        if raw_tty:
            self._restore_terminal()

        return SpawnResult(error=error, pid=pid, stdout=stdout)

    def symbol(self, symbol: int):
        """ at a symbol object from a given address """
        return Symbol.create(symbol, self)

    @property
    def errno(self) -> int:
        return self.symbols.errno[0]

    @errno.setter
    def errno(self, value: int) -> None:
        self.symbols.errno[0] = value

    @property
    def last_error(self):
        """ get info about the last occurred error """
        if not self.errno:
            return ''
        err_str = self.symbols.strerror(self.errno).peek_str()
        return f'[{self.errno}] {err_str}'

    @property
    def environ(self) -> list[str]:
        result = []
        environ = self.symbols.environ[0]
        i = 0
        while environ[i]:
            result.append(environ[i].peek_str())
            i += 1
        return result

    def setenv(self, name: str, value: str):
        """ set process environment variable """
        self.symbols.setenv(name, value)

    def getenv(self, name: str) -> typing.Optional[str]:
        """ get process environment variable """
        value = self.symbols.getenv(name)
        if not value:
            return None
        return value.peek_str()

    @property
    def pid(self):
        return int(self.symbols.getpid())

    @contextlib.contextmanager
    def safe_calloc(self, size: int):
        with self.safe_malloc(size) as x:
            x.poke(b'\x00' * size)
            try:
                yield x
            finally:
                pass

    @contextlib.contextmanager
    def safe_malloc(self, size: int):
        with self.freeing(self.symbols.malloc(size)) as x:
            yield x

    @contextlib.contextmanager
    def freeing(self, symbol):
        try:
            yield symbol
        finally:
            if symbol:
                self.symbols.free(symbol)

    def interactive(self, additional_namespace: Optional[dict] = None):
        """ Start an interactive shell """
        sys.argv = ['a']
        c = Config()
        c.IPCompleter.use_jedi = False
        c.InteractiveShellApp.exec_lines = [
            '''IPython.get_ipython().events.register('pre_run_cell', p._ipython_run_cell_hook)''',
            '''logging.getLogger('asyncio').disabled = True'''
        ]
        c.TerminalInteractiveShell.autoformatter = None
        c.BaseIPythonApplication.profile = 'rpcclient'
        namespace = globals()
        namespace.update({'p': self, 'symbols': self.symbols})
        if additional_namespace is not None:
            namespace.update(additional_namespace)
        print(USAGE)
        IPython.start_ipython(config=c, user_ns=namespace)

    def _add_global(self, name: str, value) -> None:
        globals()[name] = value

    def _ipython_run_cell_hook(self, info):
        """
        Enable lazy loading for symbols
        :param info: IPython's CellInf4o object
        """
        if not self._ipython_run_cell_hook_enabled:
            return

        if info.raw_cell.startswith('!') or info.raw_cell.endswith('?'):
            return

        for node in ast.walk(ast.parse(info.raw_cell)):
            if not isinstance(node, ast.Name):
                # we are only interested in names
                continue

            if node.id in locals() or node.id in globals() or node.id in dir(builtins):
                # That are undefined
                continue

            if not hasattr(SymbolsJar, node.id):
                # ignore SymbolsJar properties
                try:
                    symbol = getattr(self.symbols, node.id)
                except SymbolAbsentError:
                    pass
                else:
                    self._add_global(
                        node.id,
                        symbol
                    )

    def close(self):
        self._sock.close()

    def shell(self, reuse_client: bool = False):
        self._logger.disabled = True
        self._ipython_run_cell_hook_enabled = False

        args = ['--rc']
        home_rc = Path('~/.xonshrc').expanduser()
        if home_rc.exists():
            args.append(str(home_rc.expanduser().absolute()))
        args.append(str((Path(rpcclient.__file__).parent / 'xonshrc.py').absolute()))

        if reuse_client:
            XSH.ctx['_client_to_reuse'] = self
        XSH.ctx['_create_socket_cb'] = self._create_socket_cb

        try:
            logging.getLogger('parso.python.diff').disabled = True
            logging.getLogger('parso.cache').disabled = True
            logging.getLogger('asyncio').disabled = True
            xonsh_main(args)
        except SystemExit:
            self._logger.disabled = False
            self._ipython_run_cell_hook_enabled = True

    def reconnect(self):
        """ close current socket and attempt to reconnect """
        with self.reconnect_lock:
            with self._protocol_lock:
                self.close()
                self._sock = ProtoSocket(self._create_socket_cb())
            # new clients are handled in new processes so all symbols may reside in different addresses
            self._init_process_specific()

    def _execute(self, argv: list[str], envp: list[str], background=False) -> int:
        command = CmdExec(background=background, argv=argv, envp=envp)
        try:
            return self._sock.send_recv(command).pid
        except ServerResponseError:
            raise SpawnError(f'failed to spawn: {argv}')

    def _restore_terminal(self):
        if not tty_support:
            return
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, self._old_settings)

    def _prepare_terminal(self):
        if not tty_support:
            return
        fd = sys.stdin
        self._old_settings = termios.tcgetattr(fd)
        tty.setraw(fd)

    def _execution_loop(self, stdin: io_or_str = sys.stdin, stdout=sys.stdout):
        """
        if stdin is a file object, we need to select between the fds and give higher priority to stdin.
        otherwise, we can simply write all stdin contents directly to the process
        """
        fds = []
        raw_socket = self._sock.raw_socket
        if hasattr(stdin, 'fileno'):
            fds.append(stdin)
        else:
            # assume it's just raw bytes
            raw_socket.sendall(stdin.encode())
        fds.append(raw_socket)

        while True:
            response = Response()
            rlist, _, _ = select(fds, [], [])

            for fd in rlist:
                if fd == sys.stdin:
                    if stdin == sys.stdin:
                        buf = os.read(stdin.fileno(), CHUNK_SIZE)
                    else:
                        buf = stdin.read(CHUNK_SIZE)
                    raw_socket.sendall(buf)
                elif fd == raw_socket:
                    try:
                        size, buf = self._sock._receive()
                        response.ParseFromString(buf)
                    except ConnectionResetError:
                        print('Bye. 👋')
                        return
                    _type = response.exec_chunk.WhichOneof('type')
                    if _type == 'buffer':
                        stdout.write(response.exec_chunk.buffer.decode())
                        stdout.flush()
                    elif _type == 'exit_code':
                        return response.exec_chunk.exit_code

    def raise_errno_exception(self, message: str):
        message += f' ({self.last_error})'
        exceptions = {
            EPERM: RpcPermissionError,
            ENOENT: RpcFileNotFoundError,
            EEXIST: RpcFileExistsError,
            EISDIR: RpcIsADirectoryError,
            ENOTDIR: RpcNotADirectoryError,
            EPIPE: RpcBrokenPipeError,
            ENOTEMPTY: RpcNotEmptyError,
            EAGAIN: RpcResourceTemporarilyUnavailableError,
            ECONNREFUSED: RpcConnectionRefusedError,
        }
        exception = exceptions.get(self.errno)
        if exception:
            raise exception(message)
        raise BadReturnValueError(message)

    def capture_fd(self, fd: int, sock_buf_size: Optional[int] = None) -> CaptureFD:
        """
        Get a context manager, capturing output to `fd`. Read from it using the `read()` method

        sock_buf_size is required for captures above 6KB, as any write above this value would block until a read is performed.

        :param fd: FD to capture
        :param sock_buf_size: Buffer size for the capture socket, if not specified, default value is used.
        :return: CaptureFD object
        """
        return CaptureFD(self, fd, sock_buf_size)

    def __repr__(self):
        buf = f'<{self.__class__.__name__} '
        buf += f'PID:{self.symbols.getpid():d} UID:{self.symbols.getuid():d} GID:{self.symbols.getgid():d} ' \
               f'SYSNAME:{self._sysname} ARCH:{self.arch}'
        buf += '>'
        return buf



from construct import Bytes, CString, Default, FlagsEnum, FormatField, If, Int8ul, Int16sl, Int16ub, Int16ul, Int32ub, \
    Int32ul, Int64sl, Int64ul, LazyBound, PaddedString, Padding, Pointer, Struct, this

from rpcclient.structs.consts import AF_INET, AF_INET6, AF_UNIX

UNIX_PATH_MAX = 104

u_char = Int8ul
uint8_t = Int8ul
short = Int16sl
u_short = Int16ul
uint32_t = Int32ul
uint64_t = Int64ul
u_int32_t = uint32_t
uid_t = Int32ul
gid_t = Int32ul
time_t = Int32ul
long = Int64sl
mode_t = Int16ul
in_addr = Bytes(4)
in6_addr = Bytes(16)

sockaddr_in = Struct(
    'sin_family' / Default(Int16sl, AF_INET),
    'sin_port' / Int16ub,
    'sin_addr' / in_addr,
    'sin_zero' / Default(Bytes(8), b'\x00' * 8),
)

sockaddr_in6 = Struct(
    'sin6_family' / Default(Int16sl, AF_INET6),
    'sin6_port' / Int16ub,
    'sin6_flowinfo' / Default(Int32ub, 0),
    'sin6_addr' / in6_addr,
    'sin6_scope_id' / Default(Int32ub, 0),
)

sockaddr_un = Struct(
    'sun_len' / Default(Int8ul, 0),
    'sun_family' / Default(Int8ul, AF_UNIX),
    'sun_path' / PaddedString(UNIX_PATH_MAX, 'utf8'),
)

sockaddr = Struct(
    Padding(1),
    'sa_family' / Int8ul,
)

st_flags = FlagsEnum(Int32ul,
                     UF_NODUMP=1,
                     UF_IMMUTABLE=2,
                     UF_APPEND=4,
                     SF_ARCHIVED=0x00010000,
                     SF_IMMUTABLE=0x00020000,
                     SF_APPEND=0x00040000)


class SymbolFormatField(FormatField):
    """
    A Symbol wrapper for construct
    """

    def __init__(self, client):
        super().__init__('<', 'Q')
        self._client = client

    def _parse(self, stream, context, path):
        return self._client.symbol(FormatField._parse(self, stream, context, path))


def hostent(client) -> Struct:
    return Struct(
        '_h_name' / SymbolFormatField(client),
        'h_name' / Pointer(this._h_name, CString('utf8')),
        'h_aliases' / SymbolFormatField(client),
        'h_addrtype' / Int32ul,
        'h_length' / Int32ul,
        'h_addr_list' / SymbolFormatField(client),
    )


def ifaddrs(client) -> Struct:
    return Struct(
        '_ifa_next' / SymbolFormatField(client),
        'ifa_next' / If(this._ifa_next != 0, LazyBound(lambda: Pointer(this._ifa_next, ifaddrs(client)))),

        '_ifa_name' / SymbolFormatField(client),
        'ifa_name' / If(this._ifa_name != 0, Pointer(this._ifa_name, CString('utf8'))),

        'ifa_flags' / Int32ul,
        Padding(4),
        'ifa_addr' / SymbolFormatField(client),
        'ifa_netmask' / SymbolFormatField(client),
        'ifa_dstaddr' / SymbolFormatField(client),
        'ifa_data' / SymbolFormatField(client),
    )


def Dl_info(client) -> Struct:
    return Struct(
        '_dli_fname' / SymbolFormatField(client),
        'dli_fname' / Pointer(this._dli_fname, CString('utf8')),

        'dli_fbase' / SymbolFormatField(client),

        '_dli_sname' / SymbolFormatField(client),
        'dli_sname' / Pointer(this._dli_sname, CString('utf8')),

        'dli_saddr' / SymbolFormatField(client),
    )



import struct
from enum import Enum

from parameter_decorators import path_to_str

from rpcclient.allocated import Allocated
from rpcclient.darwin.consts import AVAudioSessionCategoryOptions, AVAudioSessionRouteSharingPolicy
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import BadReturnValueError, MissingLibraryError, RpcFailedToPlayError, RpcFailedToRecordError
from rpcclient.structs.consts import RTLD_NOW


class AVAudioSessionCategory(Enum):
    """ https://developer.apple.com/documentation/avfaudio/AVAudioSessionCategory?language=objc """
    PlayAndRecord = 'AVAudioSessionCategoryPlayAndRecord'
    Ambient = 'AVAudioSessionCategoryAmbient'
    MultiRoute = 'AVAudioSessionCategoryMultiRoute'
    Playback = 'AVAudioSessionCategoryPlayback'
    Record = 'AVAudioSessionCategoryRecord'
    SoloAmbient = 'AVAudioSessionCategorySoloAmbient'
    AudioProcessing = 'AVAudioSessionCategoryAudioProcessing'


class AVAudioSessionMode(Enum):
    """ https://developer.apple.com/documentation/avfaudio/avaudiosessionmode?language=objc """
    Default = 'AVAudioSessionModeDefault'
    GameChat = 'AVAudioSessionModeGameChat'
    Measurement = 'AVAudioSessionModeMeasurement'
    MoviePlayback = 'AVAudioSessionModeMoviePlayback'
    SpokenAudio = 'AVAudioSessionModeSpokenAudio'
    VideoChat = 'AVAudioSessionModeVideoChat'
    VideoRecording = 'AVAudioSessionModeVideoRecording'
    VoiceChat = 'AVAudioSessionModeVoiceChat'
    VoicePrompt = 'AVAudioSessionModeVoicePrompt'


class InterruptionPriority(Enum):
    Default = 0
    PhoneCall = 10
    EmergencyAlert = 20


class Recorder(Allocated):
    """
    Wrapper for AVAudioRecorder
    https://developer.apple.com/documentation/avfaudio/avaudiorecorder?language=objc
    """

    def __init__(self, client, session: 'AudioSession', recorder: DarwinSymbol):
        super().__init__()
        self._client = client
        self._session = session
        self._recorder = recorder

    def _deallocate(self):
        self._recorder.objc_call('release')

    def record(self):
        self._session.set_category(AVAudioSessionCategory.PlayAndRecord)
        self._session.set_active(True)
        self._recorder.objc_call('record')
        if not self.recording:
            raise RpcFailedToRecordError()

    def pause(self):
        self._recorder.objc_call('pause')
        self._session.set_active(False)

    def stop(self):
        self._recorder.objc_call('stop')
        self._session.set_active(False)

    def delete_recording(self):
        if not self._recorder.objc_call('deleteRecording'):
            raise BadReturnValueError('deleteRecording failed')

    @property
    def recording(self) -> bool:
        return bool(self._recorder.objc_call('isRecording'))


class Player(Allocated):
    """
    Wrapper for AVAudioPlayer
    https://developer.apple.com/documentation/avfaudio/avaudioplayer?language=objc
    """

    def __init__(self, client, session: 'AudioSession', player: DarwinSymbol):
        super().__init__()
        self._client = client
        self._session = session
        self._player = player

    def _deallocate(self):
        self._player.objc_call('release')

    def play(self):
        self._session.set_category(AVAudioSessionCategory.PlayAndRecord)
        self._session.set_active(True)
        self._player.objc_call('play')
        if not self.playing:
            raise RpcFailedToPlayError()

    def pause(self):
        self._player.objc_call('pause')

    def stop(self):
        self._player.objc_call('stop')
        self._session.set_active(False)

    def set_volume(self, value: float):
        self._player.objc_call('setVolume:', struct.pack('<f', value))

    @property
    def playing(self) -> bool:
        return bool(self._player.objc_call('isPlaying'))

    @property
    def loops(self) -> int:
        return self._player.objc_call('numberOfLoops')

    @loops.setter
    def loops(self, value: int):
        self._player.objc_call('setNumberOfLoops:', value)


class AudioSession:
    """
    wrapper for AVAudioSession
    https://developer.apple.com/documentation/avfaudio/avaudiosession?language=objc
    """

    def __init__(self, client):
        self._client = client
        self._session = self._client.symbols.objc_getClass('AVAudioSession').objc_call('sharedInstance')

    def set_active(self, is_active: bool):
        self._session.objc_call('setActive:error:', is_active, 0)

    def set_mode(self, mode: AVAudioSessionMode):
        mode = self._client.symbols[mode.value][0]
        self._session.objc_call('setMode:error:', mode, 0)

    def set_category(self, category: AVAudioSessionCategory,
                     mode: AVAudioSessionMode = AVAudioSessionMode.Default,
                     route_sharing_policy: AVAudioSessionRouteSharingPolicy = AVAudioSessionRouteSharingPolicy.Default,
                     options: AVAudioSessionCategoryOptions = AVAudioSessionCategoryOptions.DefaultToSpeaker) -> None:
        category = self._client.symbols[category.value][0]
        mode = self._client.symbols[mode.value][0]
        self._session.objc_call('setCategory:mode:routeSharingPolicy:options:error:', category, mode,
                                route_sharing_policy, options, 0)

    def set_interruption_priority(self, priority: InterruptionPriority):
        self._session.objc_call('setInterruptionPriority:error:', priority, 0)

    def override_output_audio_port(self, port: int):
        self._session.objc_call('overrideOutputAudioPort:error:', port, 0)

    @property
    def other_audio_playing(self) -> bool:
        return bool(self._session.objc_call('isOtherAudioPlaying'))

    @property
    def record_permission(self):
        return struct.pack('<I', self._session.objc_call('recordPermission'))[::-1].decode()

    @property
    def available_categories(self) -> list[str]:
        return self._session.objc_call('availableCategories').py()

    @property
    def available_modes(self) -> list[str]:
        return self._session.objc_call('availableModes').py()

    @property
    def is_active(self) -> bool:
        return bool(self._session.objc_call('isActive'))


class DarwinMedia:
    """ Media utils """

    def __init__(self, client):
        """
        :param rpcclient.darwin.client.DarwinClient client:
        """
        self._client = client
        self._load_av_foundation()
        self.session = AudioSession(self._client)

    def _load_av_foundation(self):
        options = [
            # macOS
            '/System/Library/Frameworks/AVFoundation.framework/Versions/A/AVFoundation',
            '/System/Library/Frameworks/AVFAudio.framework/Versions/A/AVFAudio',
            # iOS
            '/System/Library/Frameworks/AVFoundation.framework/AVFoundation'
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        raise MissingLibraryError('failed to load AVFAudio')

    @path_to_str('filename')
    def get_recorder(self, filename: str) -> Recorder:
        url = self._client.symbols.objc_getClass('NSURL').objc_call('fileURLWithPath:', self._client.cf(filename))
        settings = self._client.cf({
            'AVEncoderQualityKey': 100,
            'AVEncoderBitRateKey': 16,
            'AVNumberOfChannelsKey': 1,
            'AVSampleRateKey': 8000.0,
        })
        AVAudioRecorder = self._client.symbols.objc_getClass('AVAudioRecorder')
        recorder = AVAudioRecorder.objc_call('alloc').objc_call('initWithURL:settings:error:', url, settings, 0)

        return Recorder(self._client, self.session, recorder)

    @path_to_str('filename')
    def get_player(self, filename: str) -> Player:
        NSURL = self._client.symbols.objc_getClass('NSURL')
        url = NSURL.objc_call('fileURLWithPath:', self._client.cf(filename))

        AVAudioPlayer = self._client.symbols.objc_getClass('AVAudioPlayer')
        player = AVAudioPlayer.objc_call('alloc').objc_call('initWithContentsOfURL:error:', url, 0)

        return Player(self._client, self.session, player)



from datetime import datetime


def test_now(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.time.now() > datetime.fromtimestamp(0)


def test_set_current(client):
    """
    :param rpcclient.client.Client client:
    """
    old_now = client.time.now()
    try:
        client.time.set_current(datetime(year=2020, month=1, day=1))
        assert client.time.now() < datetime(year=2020, month=1, day=1, minute=1)
    finally:
        client.time.set_current(old_now)
        client.time.set_auto()


def test_is_set_automatically(client):
    """
    :param rpcclient.client.Client client:
    """
    client.time.set_auto()
    assert client.time.is_set_automatically



import socket
import struct
import threading

from rpcclient.exceptions import InvalidServerVersionMagicError, ServerDiedError, ServerResponseError
from rpcclient.protobuf_bridge import CmdClose, Command, Handshake, Response

# field[0] is MAGIC - skip
COMMAND_MAPPING = {field.message_type.name: field.name for field in Command.DESCRIPTOR.fields[1:]}
SERVER_MAGIC_VERSION = 0x88888809
MAGIC = 0x12345679
MAX_PATH_LEN = 1024


class ProtoSocket:
    def __init__(self, sock: socket.socket):
        self.handshake = None
        self.raw_socket = sock
        self._protocol_lock = threading.Lock()
        self._do_handshake()

    def _do_handshake(self) -> None:
        self.handshake = Handshake()
        size, buff = self._receive()
        self.handshake.ParseFromString(buff)
        if self.handshake.magic != SERVER_MAGIC_VERSION:
            raise InvalidServerVersionMagicError(f'got {self.handshake.magic:x} instead of {SERVER_MAGIC_VERSION:x}')

    def send_recv(self, sub_command):
        command = Command(magic=MAGIC, **{COMMAND_MAPPING.get(sub_command.DESCRIPTOR.name): sub_command})
        response = Response()
        with self._protocol_lock:
            self._send(command.SerializeToString())
            size, buff = self._receive()
        command_type = command.WhichOneof('type')
        response.ParseFromString(buff)

        # Iterate through all possible response types
        if response.HasField('error'):
            raise ServerResponseError(response.error)
        else:
            return getattr(response, command_type.lower())

    def _receive(self) -> tuple[int, bytes]:
        try:
            size = struct.unpack('<Q', self.raw_socket.recv(8))[0]
            buff = self._recvall(size)
            return size, buff
        except struct.error:
            raise ConnectionError()

    def _send(self, message: bytes) -> None:
        buff = struct.pack('<Q', len(message)) + message
        self.raw_socket.sendall(buff)

    def _recvall(self, size: int) -> bytes:
        buf = b''
        while size:
            try:
                chunk = self.raw_socket.recv(size)
            except BlockingIOError:
                continue
            if not self.raw_socket.getblocking() and not chunk:
                raise ServerDiedError()
            size -= len(chunk)
            buf += chunk
        return buf

    def close(self) -> None:
        command = Command(magic=MAGIC, close=CmdClose())
        self._send(command.SerializeToString())
        self.raw_socket.close()



import pytest

from rpcclient.darwin.consts import IOPMUserActiveType

pytestmark = pytest.mark.darwin


def test_copy_assertions_status(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    assertions = client.power.copy_assertions_status()
    assert len(assertions) > 0


def test_copy_assertions_by_process(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    client.power.copy_assertions_by_process()


def test_declare_user_activity(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    name = 'test user activity'
    with client.power.declare_user_activity(name, IOPMUserActiveType.kIOPMUserActiveLocal):
        assert client.power.copy_assertions_by_process()[client.pid][0]['AssertName'] == name


def test_declare_user_network_activity(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    name = 'test user network activity'
    with client.power.declare_network_client_activity(name):
        assert client.power.copy_assertions_by_process()[client.pid][0]['AssertName'] == name


def test_create_assertion(client):
    """
    :param rpcclient.darwin.client.DarwinClient client:
    """
    name = 'test assertion'
    with client.power.create_assertion(name, 'PreventUserIdleSystemSleep'):
        assert client.power.copy_assertions_by_process()[client.pid][0]['AssertName'] == name



import platform
import plistlib
import posixpath
import uuid
from datetime import datetime, timedelta

from rpcclient.darwin.scpreferences import SCPreference

PAIR_RECORD_PATH = '/var/root/Library/Lockdown/pair_records'
DATA_ARK_PATH = '/var/root/Library/Lockdown/data_ark.plist'
FAR_FUTURE_DATE = datetime(9999, 1, 1)


class PairRecord:
    def __init__(self, client, host_id: str):
        self._client = client
        self._host_id = host_id

    @property
    def host_id(self) -> str:
        return self._host_id

    @property
    def record(self) -> dict:
        return plistlib.loads(self._client.fs.read_file(posixpath.join(PAIR_RECORD_PATH, f'{self._host_id}.plist')))

    @property
    def date(self) -> datetime:
        return self._client.lockdown.pair_dates.get(self._host_id)

    @date.setter
    def date(self, value: datetime):
        self._client.lockdown.set_pair_date(self._host_id, value)

    @property
    def expiration_date(self) -> datetime:
        return self.date + timedelta(days=30)

    def disable_expiration(self):
        self.date = FAR_FUTURE_DATE

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} HOST_ID:{self.host_id} EXPIRATION:{self.expiration_date}>'


class Lockdown:
    def __init__(self, client):
        self._client = client

    @staticmethod
    def get_host_id(hostname: str = None) -> str:
        hostname = platform.node() if hostname is None else hostname
        host_id = uuid.uuid3(uuid.NAMESPACE_DNS, hostname)
        return str(host_id).upper()

    @property
    def pair_records(self) -> list[PairRecord]:
        """ list pair records """
        result = []
        for filename in self._client.fs.listdir(PAIR_RECORD_PATH):
            result.append(PairRecord(self._client, filename.split('.')[0]))
        return result

    @property
    def pair_dates(self) -> dict:
        result = {}
        raw = self._client.preferences.cf.get_dict('com.apple.mobile.ldpair', 'mobile', 'kCFPreferencesAnyHost')
        for host_id, timestmap in raw.items():
            result[host_id] = datetime.fromtimestamp(timestmap)
        return result

    @property
    def data_ark(self) -> SCPreference:
        return self._client.preferences.sc.open(DATA_ARK_PATH)

    def set_pair_date(self, host_id: str, date: datetime):
        self._client.preferences.cf.set(host_id, int(date.timestamp()), 'com.apple.mobile.ldpair', 'mobile',
                                        'kCFPreferencesAnyHost')

    def get_pair_record_by_host_id(self, host_id: str) -> PairRecord:
        return PairRecord(self._client, host_id)

    def get_pair_record_by_hostname(self, hostname: str) -> PairRecord:
        return PairRecord(self._client, self.get_host_id(hostname))

    def get_self_pair_record(self) -> PairRecord:
        return self.get_pair_record_by_host_id(self.get_host_id())

    def add_pair_record(self, pair_record: dict, date: datetime, hostname: str = None):
        pair_record = dict(pair_record)
        # remove private key from pair record before adding it
        pair_record.pop('HostPrivateKey')

        host_id = self.get_host_id(hostname)
        self._client.fs.write_file(posixpath.join(PAIR_RECORD_PATH, f'{host_id}.plist'), plistlib.dumps(pair_record))
        self.set_pair_date(host_id, date)

    def disable_expiration_for_all_existing_pair_records(self):
        for record in self.pair_records:
            record.disable_expiration()



import pytest

STDOUT_FD = 1
STDERR_FD = 2
TEST_DATA: bytes = b'testtesttest'


@pytest.mark.parametrize('fd', [
    STDOUT_FD,
    STDERR_FD
])
def test_capture_fd(client, fd: int) -> None:
    """
    :param rpcclient.client.Client client:
    :param int fd:
    """
    with client.capture_fd(fd) as cap:
        if -1 == client.symbols.write(fd, TEST_DATA, len(TEST_DATA)):
            client.raise_errno_exception('write to fd failed')
        data = cap.read()
    assert TEST_DATA == data


@pytest.mark.parametrize('fd', [
    STDOUT_FD,
    STDERR_FD
])
def test_fd_cleanup(client, fd: int) -> None:
    """
    :param rpcclient.client.Client client:
    :param int fd:
    """
    fd_count = len(client.processes.get_self().fds)
    with client.capture_fd(fd):
        pass
    assert fd_count == len(client.processes.get_self().fds)



from typing import Optional

from osstatus.cache import ErrorCode, get_possible_error_codes

from rpcclient.exceptions import UnrecognizedSelectorError
from rpcclient.symbol import Symbol


class DarwinSymbol(Symbol):
    def objc_call(self, selector, *params, **kwargs):
        """ call an objc method on a given object """
        sel = self._client.symbols.sel_getUid(selector)
        if not self._client.symbols.objc_msgSend(self, self._client.symbols.sel_getUid("respondsToSelector:"), sel):
            raise UnrecognizedSelectorError(f"unrecognized selector '{selector}' sent to class")

        return self._client.symbols.objc_msgSend(self, sel, *params, **kwargs)

    def py(self, *args, **kwargs):
        """ get a python object from a core foundation one """
        if self == 0:
            return None

        return self._client.decode_cf(self)

    @property
    def region(self):
        """ get corresponding region """
        for region in self._client.processes.get_by_pid(self._client.pid).regions:
            if (self >= region.start) and (self <= region.end):
                return region

    @property
    def cfdesc(self):
        """
        Get output from CFCopyDescription()
        :return: CFCopyDescription()'s output as a string
        """
        if self == 0:
            return None
        return self._client.symbols.CFCopyDescription(self).py()

    @property
    def objc_symbol(self):
        """
        Get an ObjectiveC symbol of the same address
        :return: Object representing the ObjectiveC symbol
        """
        return self._client.objc_symbol(self)

    @property
    def osstatus(self) -> Optional[list[ErrorCode]]:
        """ Get possible translation to given error code by querying osstatus """
        return get_possible_error_codes(self)

    @property
    def stripped_value(self) -> 'DarwinSymbol':
        """ Remove PAC upper bits """
        return self._client.symbol(self & 0xfffffffff)



import typing

from rpcclient.darwin.client import DarwinClient
from rpcclient.darwin.reports import Reports
from rpcclient.macos.apple_script import AppleScript

CRASH_REPORTS_DIR = 'Library/Logs/DiagnosticReports'


class MacosClient(DarwinClient):
    def __init__(self, sock, sysname: str, arch, create_socket_cb: typing.Callable):
        super().__init__(sock, sysname, arch, create_socket_cb)
        self.reports = Reports(self, CRASH_REPORTS_DIR)
        self.apple_script = AppleScript(self)

    @property
    def roots(self) -> list[str]:
        """ get a list of all accessible darwin roots when used for lookup of files/preferences/... """

        result = super().roots
        for username in self.fs.scandir('/Users'):
            if not username.is_dir() or not self.fs.accessible(username.path):
                continue
            result.append(username.path)
        return result



import logging

from rpcclient.exceptions import RpcSetDeveloperModeError
from rpcclient.structs.consts import RTLD_NOW

logger = logging.getLogger(__name__)


class Amfi:
    """ AMFI utils """

    def __init__(self, client):
        self._client = client
        self._load_amfi()

    def set_developer_mode_status(self, enabled: bool) -> None:
        cfreply = self._client.xpc.send_message_using_cf_serialization(
            'com.apple.amfi.xpc', {'action': int(not (enabled))}, False)['cfreply']
        raw_response = list(cfreply.values())[0]
        if b'success' not in raw_response:
            raise RpcSetDeveloperModeError()

    @property
    def developer_mode_status(self) -> bool:
        """ get Developer Mode status """
        return self._client.symbols.amfi_developer_mode_status()

    def _load_amfi(self):
        options = [
            '/System/Library/PrivateFrameworks/AppleMobileFileIntegrity.framework/AppleMobileFileIntegrity',
        ]
        for option in options:
            if self._client.dlopen(option, RTLD_NOW):
                return
        logger.warning('AppleMobileFileIntegrity library isn\'t available')



from collections import namedtuple

from rpcclient.darwin.symbol import DarwinSymbol

CGRect = namedtuple('CGRect', 'x0 y0 x1 y1')


class ScreenCapture:
    """ monitor screen events """

    def __init__(self, client):
        self._client = client

    @property
    def main_display(self) -> DarwinSymbol:
        return self._client.symbols.objc_getClass('CADisplay').objc_call('mainDisplay')

    @property
    def bounds(self) -> CGRect:
        result = self.main_display.objc_call('bounds', return_raw=True)
        return CGRect(x0=result.d0, y0=result.d1, x1=result.d2, y1=result.d3)

    @property
    def screenshot(self) -> bytes:
        return self._client.symbols.UIImagePNGRepresentation(self._client.symbols._UICreateScreenUIImage()).py()



import contextlib
import logging
import os
import posixpath
import stat
import tempfile
from collections.abc import Generator
from pathlib import Path, PosixPath
from random import Random
from typing import Union

from parameter_decorators import path_to_str

from rpcclient.allocated import Allocated
from rpcclient.darwin.structs import MAXPATHLEN
from rpcclient.darwin.symbol import DarwinSymbol
from rpcclient.exceptions import ArgumentError, BadReturnValueError, RpcFileExistsError, RpcFileNotFoundError, \
    RpcIsADirectoryError
from rpcclient.structs.consts import DT_DIR, DT_LNK, DT_REG, DT_UNKNOWN, O_CREAT, O_RDONLY, O_RDWR, O_TRUNC, O_WRONLY, \
    R_OK, S_IFDIR, S_IFLNK, S_IFMT, S_IFREG, SEEK_CUR

logger = logging.getLogger(__name__)


class DirEntry:
    def __init__(self, path, entry, client):
        self._path = path
        self._entry = entry
        self._client = client
        self._lstat = None
        self._stat = None

    @property
    def name(self) -> str:
        return self._entry.d_name

    @property
    def path(self) -> str:
        return posixpath.join(self._path, self.name)

    def inode(self):
        """ Return inode of the entry; cached per entry. """
        return self._entry.d_ino

    def is_dir(self, *, follow_symlinks=True):
        """ Return True if the entry is a directory; cached per entry. """
        return self._test_mode(follow_symlinks, S_IFDIR)

    def is_file(self, *, follow_symlinks=True):
        """ Return True if the entry is a file; cached per entry. """
        return self._test_mode(follow_symlinks, S_IFREG)

    def is_symlink(self):
        """ Return True if the entry is a symbolic link; cached per entry. """
        if self._entry.d_type != DT_UNKNOWN:
            return self._entry.d_type == DT_LNK
        else:
            return self._test_mode(False, S_IFLNK)

    def stat(self, *, follow_symlinks=True):
        """ Return stat_result object for the entry; cached per entry. """
        if not follow_symlinks:
            return self._get_lstat()
        if self._stat is None:
            if self.is_symlink():
                self._stat = self._fetch_stat(True)
            else:
                self._stat = self._get_lstat()
        return self._stat

    def _test_mode(self, follow_symlinks, mode):
        is_symlink = self._entry.d_type == DT_LNK
        need_stat = self._entry.d_type == DT_UNKNOWN or (follow_symlinks and is_symlink)
        if not need_stat:
            d_types = {S_IFLNK: DT_LNK, S_IFDIR: DT_DIR, S_IFREG: DT_REG}
            return self._entry.d_type == d_types[mode]
        else:
            st_mode = self.stat(follow_symlinks=follow_symlinks).st_mode
            if not st_mode:
                self._client.raise_errno_exception(f'failed to stat(): {self.path}')
            return (st_mode & S_IFMT) == mode

    def _get_lstat(self):
        if self._lstat is None:
            self._lstat = self._fetch_stat(False)
        return self._lstat

    def _fetch_stat(self, follow_symlinks):
        result = self._entry.stat
        if not follow_symlinks:
            result = self._entry.lstat

        if result.errno != 0:
            self._client.errno = result.errno
            self._client.raise_errno_exception(f'failed to stat: {self._entry.d_name}')
        return result

    def __repr__(self):
        return f'<{self.__class__.__name__} NAME:{self.name}{"/" if self.is_dir() else ""}>'


class File(Allocated):
    CHUNK_SIZE = 1024 * 64

    def __init__(self, client, fd: int):
        """
        :param rpcclient.client.client.Client client:
        :param fd:
        """
        super().__init__()
        self._client = client
        self.fd = fd

    def _deallocate(self):
        """ close(fd) at remote. read man for more details. """
        fd = self._client.symbols.close(self.fd).c_int32
        if fd < 0:
            self._client.raise_errno_exception(f'failed to close fd: {fd}')

    def seek(self, offset: int, whence: int) -> int:
        """ lseek(fd, offset, whence) at remote. read man for more details. """
        err = self._client.symbols.lseek(self.fd, offset, whence).c_int32
        if err < 0:
            self._client.raise_errno_exception(f'failed to lseek fd: {self.fd}')
        return err

    def tell(self) -> int:
        return self.seek(0, SEEK_CUR)

    def _write(self, buf: bytes) -> int:
        """ write(fd, buf, size) at remote. read man for more details. """
        n = self._client.symbols.write(self.fd, buf, len(buf)).c_int64
        if n < 0:
            self._client.raise_errno_exception(f'failed to write on fd: {self.fd}')
        return n

    def write(self, buf: bytes) -> int:
        """ continue call write() until """
        while buf:
            err = self._write(buf)
            buf = buf[err:]
        return len(buf)

    def _read(self, buf: DarwinSymbol, size: int) -> bytes:
        """ read file at remote """
        err = self._client.symbols.read(self.fd, buf, size).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'read() failed for fd: {self.fd}')
        return buf.peek(err)

    def read_using_chunk(self, chunk: DarwinSymbol, chunk_size: int, size: int) -> bytes:
        buf = b''
        while size == -1 or len(buf) < size:
            read_chunk = self._read(chunk, chunk_size)
            if not read_chunk:
                # EOF
                break
            buf += read_chunk
        return buf

    def read(self, size: int = -1, chunk_size: int = CHUNK_SIZE, chunk: DarwinSymbol = None) -> bytes:
        """ read file at remote """
        if size != -1 and size < chunk_size:
            chunk_size = size

        buf = b''
        if chunk:
            return self.read_using_chunk(chunk, chunk_size, size)
        else:
            with self._client.safe_malloc(chunk_size) as temp_chunk:
                return self.read_using_chunk(temp_chunk, chunk_size, size)
        return buf

    def pread(self, length: int, offset: int) -> bytes:
        """ call pread() at remote """
        with self._client.safe_malloc(length) as buf:
            err = self._client.symbols.pread(self.fd, buf, length, offset).c_int64
            if err < 0:
                self._client.raise_errno_exception(f'pread() failed for fd: {self.fd}')
            return buf.peek(err)

    def pwrite(self, buf: bytes, offset: int):
        """ call pwrite() at remote """
        err = self._client.symbols.pwrite(self.fd, buf, len(buf), offset).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'pwrite() failed for fd: {self.fd}')

    def fdatasync(self):
        err = self._client.symbols.fdatasync(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'fdatasync() failed for fd: {self.fd}')

    def fsync(self):
        err = self._client.symbols.fsync(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'fsync() failed for fd: {self.fd}')

    def dup(self) -> int:
        err = self._client.symbols.dup(self.fd).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'dup() failed for fd: {self.fd}')
        return err

    def __repr__(self):
        return f'<{self.__class__.__name__} FD:{self.fd}>'


class RemotePath(PosixPath):
    def __init__(self, path: str, client) -> None:
        try:
            super().__init__(path)  # solution from python3.12 since signature has changed
        except TypeError:
            super().__init__()
        self._path = path
        self._client = client

    def __new__(cls, path: str, client):
        # this will not be needed once python3.11 is deprecated since it is now possible to subclass normally
        return super().__new__(cls, *[path])

    def chmod(self, mode: int):
        return self._client.fs.chmod(self._path, mode)

    def readlink(self):
        return self._client.fs.readlink(self._path)

    def exists(self) -> bool:
        try:
            self.stat()
            return True
        except Exception:
            return False

    def is_dir(self) -> bool:
        return bool(self.stat().st_mode & S_IFDIR)

    def lstat(self):
        return self._client.fs.lstat(self._path)

    def mkdir(self, mode: int, exist_ok=False):
        self._client.fs.mkdir(self._path, mode, exist_ok)

    def read_bytes(self) -> bytes:
        with self._open('r') as f:
            return f.read()

    def stat(self):
        return self._client.fs.stat(self._path)

    def symlink_to(self, target: Path, target_is_directory: bool = False) -> None:
        return self._client.fs.symlink(target, self._path)

    def write_bytes(self, buf: bytes) -> int:
        with self._open('w') as f:
            return f.write(buf)

    def _open(self, mode: str, access: int = 0o777) -> File:
        return self._client.fs.open(self._path, mode, access)

    def iterdir(self) -> Generator['RemotePath', None, None]:
        for entry in self._client.fs.listdir(self._path):
            yield self.__class__(f'{self._path}/{entry}', self._client)

    def touch(self, mode: int = 438, exist_ok: bool = True) -> None:
        try:
            return self._client.fs.touch(self._path, mode, exist_ok)
        except RpcFileExistsError:
            raise FileExistsError()

    def __truediv__(self, key: Path) -> 'RemotePath':
        return RemotePath(f'{self._path}/{key}', self._client)

    def remove(self, recursive: bool = False, force: bool = False):
        """Remove the current path from the filesystem.

        :param recursive:
        :param force:
        """
        if self.exists():
            self._client.fs.remove(self._path, recursive=recursive, force=force)


class RemoteTemporaryDir(Allocated, RemotePath):
    """ Temporary directory on the remote device """

    def __init__(self, client, directory: str = '/tmp', mode: int = 0o700):
        """Generate a random temp directory name and create it in the given directory (default '/tmp').

        :param rpcclient.darwin.client.DarwinClient client:
        :param directory: Directory to create the temp directory inside (default '/tmp').
        :param mode: Mode to create the temp directory with (default 0o700).
        """
        Allocated.__init__(self)
        remote_dir = client.fs.remote_path(directory)
        if not remote_dir.exists:
            raise RpcFileNotFoundError(f'remote dir {remote_dir} does not exist')
        name = 'tmpdir_' + ''.join(Random().choices('abcdefghijklmnopqrstuvwxyz0123456789_', k=8))
        remote_path = client.fs.remote_path(str(remote_dir / Path(name)))
        RemotePath.__init__(self, remote_path, client)
        self.mkdir(mode)

    def __new__(cls, *args, **kwargs):
        # this will not be needed once python3.11 is deprecated since it is now possible to subclass normally
        return RemotePath.__new__(cls, *[args[1], args[0]])

    def _deallocate(self):
        """Remove the temp directory from the filesystem recursively"""
        self.remove(recursive=True, force=True)


class Fs:
    """ filesystem utils """

    def __init__(self, client):
        self._client = client

    def _cp_dir(self, source: Path, dest: Path, force: bool):
        if not dest.exists():
            dest.mkdir(source.lstat().st_mode)

        files = source.iterdir()

        for src_file in files:
            dest_file = dest / src_file.name

            src_lstat = src_file.lstat()
            if stat.S_ISDIR(src_lstat.st_mode):
                self._cp_dir(src_file, dest_file, force)
            elif stat.S_ISLNK(src_lstat.st_mode):
                symlink_full = src_file.readlink()
                dest_file.symlink_to(symlink_full)
            elif dest_file.exists() and not force:
                pass
            else:
                dest_file.write_bytes(src_file.read_bytes())

    def cp(self, sources: list[Path], dest: Path, recursive: bool, force: bool):
        dest_exists = dest.exists()
        is_dest_dir = dest_exists and dest.is_dir()

        if (not dest_exists or not is_dest_dir) and (len(sources) > 1):
            raise ArgumentError(f'target {dest} is not a directory')

        if recursive:
            if not dest_exists:
                try:
                    dest.mkdir(0o777)
                except Exception as e:
                    if not dest.exists():
                        raise e

        for source in sources:
            if not source.exists():
                raise ArgumentError(f'cannot stat {source}: No such file or directory')

            if source.is_dir():
                if not recursive:
                    logger.info(f'omitting directory {source}')
                else:
                    cur_dest = dest / source.name
                    source_mode = source.stat().st_mode
                    if not cur_dest.exists():
                        cur_dest.mkdir(source_mode)
                    self._cp_dir(source, cur_dest, force)
            else:  # source is a file
                cur_dest = dest
                if dest.exists() and dest.is_dir():
                    cur_dest = dest / source.name
                if not cur_dest.exists() or force:
                    cur_dest.write_bytes(source.read_bytes())

    @path_to_str('path')
    def is_file(self, path: str) -> bool:
        """ Return True if the entry is a file """
        return bool(self.stat(path).st_mode & S_IFREG)

    @path_to_str('path')
    def _chown(self, path: str, uid: int, gid: int):
        """ chmod(path, mode) at remote. read man for more details. """
        if self._client.symbols.chown(path, uid, gid).c_int32 < 0:
            self._client.raise_errno_exception(f'failed to chown: {path}')

    @path_to_str('path')
    def chown(self, path: str, uid: int, gid: int, recursive=False):
        """ chmod(path, mode) at remote. read man for more details. """
        if not recursive:
            self._chown(path, uid, gid)
            return

        for file in self.find(path, topdown=False):
            self._chown(file, uid, gid)

    @path_to_str('path')
    def _chmod(self, path: str, mode: int):
        """ chmod(path, mode) at remote. read man for more details. """
        if self._client.symbols.chmod(path, mode).c_int32 < 0:
            self._client.raise_errno_exception(f'failed to chmod: {path}')

    @path_to_str('path')
    def chmod(self, path: str, mode: int, recursive=False):
        """ chmod(path, mode) at remote. read man for more details. """
        if not recursive:
            self._chmod(path, mode)
            return

        for file in self.find(path, topdown=False):
            self._chmod(file, mode)

    @path_to_str('path')
    def _remove(self, path: str, force=False):
        """ remove(path) at remote. read man for more details. """
        if self._client.symbols.remove(path).c_int32 < 0:
            if not force:
                self._client.raise_errno_exception(f'failed to remove: {path}')

    @path_to_str('path')
    def remove(self, path: str, recursive=False, force=False):
        """ remove(path) at remote. read man for more details. """
        if not recursive or self.is_file(path):
            self._remove(path, force=force)
            return

        for filename in self.find(path, topdown=False):
            self._remove(filename, force=force)

    @path_to_str('old')
    @path_to_str('new')
    def rename(self, old: str, new: str):
        """ rename(old, new) at remote. read man for more details. """
        if self._client.symbols.rename(old, new).c_int32 < 0:
            self._client.raise_errno_exception(f'failed to rename: {old} -> {new}')

    @path_to_str('path')
    def _mkdir(self, path: str, mode: int = 0o777):
        """ mkdir(path, mode) at remote. read man for more details. """
        if self._client.symbols.mkdir(path, mode).c_int64 < 0:
            self._client.raise_errno_exception(f'failed to mkdir: {path}')

        # os may not always respect the permission given by the mode argument to mkdir
        self.chmod(path, mode)

    @path_to_str('path')
    def mkdir(self, path: str, mode: int = 0o777, parents=False, exist_ok=False):
        """ mkdir(path, mode) at remote. read man for more details. """
        if not parents:
            try:
                self._mkdir(path, mode=mode)
            except RpcFileExistsError:
                if not exist_ok:
                    raise
            return

        dir_path = Path(self.pwd())
        for part in Path(path).parts:
            dir_path = dir_path / part
            try:
                self._mkdir(dir_path, mode=mode)
            except (RpcIsADirectoryError, RpcFileExistsError):
                pass

    @path_to_str('path')
    def chdir(self, path: str):
        """ chdir(path) at remote. read man for more details. """
        if self._client.symbols.chdir(path).c_int64 < 0:
            self._client.raise_errno_exception(f'failed to chdir: {path}')

    @path_to_str('path')
    def readlink(self, path: str, absolute=True) -> str:
        """ readlink() at remote. read man for more details. """
        with self._client.safe_calloc(MAXPATHLEN) as buf:
            if self._client.symbols.readlink(path, buf, MAXPATHLEN).c_int64 < 0:
                self._client.raise_errno_exception(f'readlink failed for: {path}')
            if absolute:
                return str(Path(path).parent / buf.peek_str())
            return buf.peek_str()

    @path_to_str('path')
    def realpath(self, path: str) -> str:
        """ realpath() at remote. read man for more details. """
        with self._client.safe_malloc(MAXPATHLEN) as buf:
            if self._client.symbols.realpath(path, buf) == 0:
                self._client.raise_errno_exception(f'realpath failed for: {path}')
            return buf.peek_str()

    @path_to_str('path')
    def is_symlink(self, path: str) -> bool:
        try:
            self.readlink(path)
            return True
        except BadReturnValueError:
            return False

    @path_to_str('file')
    def open(self, file: str, mode: str, access: int = 0o777) -> File:
        """
        call open(file, mode, access) at remote and get a context manager
        file object
        :param file: filename to be opened
        :param mode: one of:
            'r' - read only
            'r+' - read and write. exception if file doesn't exist
            'rw' - read and write. create if it doesn't exist. also truncate.
            'w' - write only. create if it doesn't exist. also truncate.
            'w+' - read and write. create if it doesn't exist.
        :param access: access mode as octal value
        :return: a context manager file object
        """
        available_modes = {
            'r': O_RDONLY,
            'r+': O_RDWR,
            'rw': O_RDWR | O_CREAT | O_TRUNC,
            'w': O_WRONLY | O_CREAT | O_TRUNC,
            'w+': O_RDWR | O_CREAT,
        }
        mode_int = available_modes.get(mode)
        if mode_int is None:
            raise ArgumentError(f'mode can be only one of: {available_modes.keys()}')

        fd = self._client.symbols.open(file, mode_int, access, va_list_index=2).c_int32
        if fd < 0:
            self._client.raise_errno_exception(f'failed to open: {file}')

        return File(self._client, fd)

    @path_to_str('file')
    def write_file(self, file: str, buf: bytes, access: int = 0o777):
        with self.open(file, 'w', access=access) as f:
            f.write(buf)

    @path_to_str('file')
    def read_file(self, file: str) -> bytes:
        with self.open(file, 'r') as f:
            return f.read()

    def remote_path(self, path: str) -> RemotePath:
        return RemotePath(path, self._client)

    @path_to_str('local')
    def pull(self, remotes: Union[list[Union[str, Path]], Union[str, Path]], local: str, recursive: bool = False,
             force: bool = False):
        """ pull complete directory tree """
        if not isinstance(remotes, list):
            remotes = [posixpath.expanduser(remotes)]
        remotes_str = [posixpath.expanduser(remote) for remote in remotes]
        self.cp([self.remote_path(remote) for remote in remotes_str], Path(str(local)), recursive, force)

    @path_to_str('remote')
    def push(self, locals: Union[list[Union[str, Path]], Union[str, Path]], remote: str, recursive: bool = False,
             force: bool = False):
        """ push complete directory tree """
        if not isinstance(locals, list):
            locals = [posixpath.expanduser(locals)]
        locals_str = [posixpath.expanduser(local) for local in locals]
        self.cp([Path(str(local)) for local in locals_str], self.remote_path(remote), recursive, force)

    @path_to_str('file')
    def touch(self, file: str, mode: int = 0o666, exist_ok=True):
        """ simulate unix touch command for given file """
        if not exist_ok:
            try:
                self.stat(file)
                raise RpcFileExistsError()
            except RpcFileNotFoundError:
                pass
        with self.open(file, 'w+', mode):
            pass

    @path_to_str('src', 'dst')
    def symlink(self, src: str, dst: str) -> int:
        """ symlink(src, dst) at remote. read man for more details. """
        err = self._client.symbols.symlink(src, dst).c_int64
        if err < 0:
            self._client.raise_errno_exception(
                f'symlink failed to create link: {dst}->{src}')
        return err

    @path_to_str('src', 'dst')
    def link(self, src: str, dst: str) -> int:
        """ link(src, dst) - hardlink at remote. read man for more details. """
        err = self._client.symbols.link(src, dst).c_int64
        if err < 0:
            self._client.raise_errno_exception(f'link failed to create link: {dst}->{src}')
        return err

    def pwd(self) -> str:
        """ calls getcwd(buf, size_t) and prints current path.
            with the special values NULL, 0 the buffer is allocated dynamically """
        chunk = self._client.symbols.getcwd(0, 0)
        if chunk == 0:
            self._client.raise_errno_exception('getcwd() failed')
        buf = chunk.peek_str()
        self._client.symbols.free(chunk)
        return buf

    @path_to_str('path')
    def listdir(self, path: str = '.') -> list[str]:
        """ get directory listing for a given dirname """
        return [e.name for e in self.scandir(path)]

    @path_to_str('path')
    def scandir(self, path: str = '.') -> list[DirEntry]:
        """ get directory listing for a given dirname """
        result = []
        for entry in self._client.listdir(path)[2:]:
            result.append(DirEntry(path, entry, self._client))
        return result

    @path_to_str('path')
    def stat(self, path: str):
        """ stat(filename) at remote. read man for more details. """
        raise NotImplementedError()

    @path_to_str('path')
    def lstat(self, path: str):
        """ lstat(filename) at remote. read man for more details. """
        raise NotImplementedError()

    @path_to_str('path')
    def accessible(self, path: str, mode: int = R_OK):
        """ check if a given path can be accessed. """
        err = self._client.symbols.access(path, mode)
        return err == 0

    @path_to_str('path')
    def chflags(self, path: str, flags: int = 0):
        """ set file flags """
        err = self._client.symbols.chflags(path, flags)
        if err < 0:
            self._client.raise_errno_exception(f'failed to chflags on: {path}')

    @path_to_str('top')
    def find(self, top: str, topdown=True):
        """ traverse a file tree top to down """
        if topdown:
            if self.accessible(top):
                yield top
            else:
                raise RpcFileNotFoundError(f'cannot access: {top}')

        for root, dirs, files in self.walk(top, topdown=topdown):
            for name in files:
                yield os.path.join(root, name)
            for name in dirs:
                yield os.path.join(root, name)

        if not topdown:
            if self.accessible(top):
                yield top
            else:
                raise RpcFileNotFoundError(f'cannot access: {top}')

    @path_to_str('top')
    def walk(self, top: str, topdown=True, onerror=None):
        """ provides the same results as os.walk(top) """
        dirs = []
        files = []
        try:
            for entry in self.scandir(top):
                try:
                    if entry.is_dir():
                        dirs.append(entry.name)
                    else:
                        files.append(entry.name)
                except Exception as e:
                    if not onerror:
                        raise e
                    onerror(e)
        except Exception as e:
            if not onerror:
                raise e
            onerror(e)

        if topdown:
            yield top, dirs, files

        if dirs:
            for d in dirs:
                yield from self.walk(posixpath.join(top, d), topdown=topdown, onerror=onerror)

        if not topdown:
            yield top, dirs, files

    @contextlib.contextmanager
    @path_to_str('remote')
    def remote_file(self, remote: str):
        with tempfile.TemporaryDirectory() as local_dir:
            local = Path(local_dir) / Path(remote).parts[-1]
            if self.accessible(remote):
                self.pull(remote, local.absolute())
            try:
                yield local.absolute()
            finally:
                self.push(local, remote, force=True)

    @path_to_str('directory')
    def remote_temp_dir(self, directory: str = '/tmp', mode: int = 0o700) -> RemoteTemporaryDir:
        """Generate a random temp directory name and create it in the given directory (default '/tmp').

        :param directory: Directory to create the temp directory inside (default '/tmp').
        :param mode: Mode to create the temp directory with (default 0o700).
        """
        return RemoteTemporaryDir(self._client, directory, mode)



def test_valid_gethostbyname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.network.gethostbyname('google.com') is not None


def test_invalid_gethostbyname(client):
    """
    :param rpcclient.client.Client client:
    """
    assert client.network.gethostbyname('google.com1') is None
